from __future__ import annotations

import time
from argparse import Namespace
from functools import partial
from typing import Any, cast

import attr
from blessed import Terminal

from . import __version__, activities, handlers, keys, types, utils, views, widgets
from .config import Configuration, Flag
from .data import Data


def main(
    term: Terminal,
    config: Configuration | None,
    data: Data,
    host: types.Host,
    options: Namespace,
    *,
    render_header: bool = True,
    render_footer: bool = True,
    width: int | None = None,
    wait_on_actions: float | None = None,
) -> None:
    fs_blocksize = options.blocksize

    is_local = data.pg_is_local() and data.pg_is_local_access()

    skip_db_size = not options.dbsize
    server_information = data.pg_get_server_information(
        prev_server_info=None,
        using_rds=options.rds,
        skip_db_size=skip_db_size,
        skip_tempfile=not options.tempfiles,
        skip_walreceiver=not options.walreceiver,
    )

    flag = Flag.load(config, is_local=is_local, **vars(options))
    ui = types.UI.make(
        header=types.UIHeader.make(
            config.header() if config else None,
            show_instance=options.header_show_instance,
            show_system=options.header_show_system,
            show_workers=options.header_show_workers,
        ),
        config=config,
        flag=flag,
        refresh_time=options.refresh,
        min_duration=options.minduration,
        duration_mode=int(options.durationmode),
        wrap_query=options.wrap_query,
        max_db_length=min(max(server_information.max_dbname_length, 8), 16),
        filters=data.filters,
    )

    key, in_help = None, False
    sys_procs: dict[int, types.SystemProcess] = {}
    pg_procs = types.SelectableProcesses([])
    activity_stats: types.ActivityStats

    msg_pile = utils.MessagePile(2)

    with term.fullscreen(), term.cbreak(), term.hidden_cursor():
        while True:
            if key == keys.HELP:
                in_help = True
            elif in_help and key is not None:
                in_help, key = False, None
                print(term.clear + term.home, end="")
            elif key == keys.EXIT:
                break
            elif not ui.interactive() and key == keys.SPACE:
                ui.toggle_pause()
            elif not options.dbsize and key == keys.REFRESH_DB_SIZE:
                skip_db_size = False
            elif key is not None:
                if keys.is_process_next(key):
                    if pg_procs.focus_next():
                        ui.start_interactive()
                elif keys.is_process_prev(key):
                    if pg_procs.focus_prev():
                        ui.start_interactive()
                elif keys.is_process_nextpage(key):
                    if pg_procs.focus_next(term.height // 3):
                        ui.start_interactive()
                elif keys.is_process_prevpage(key):
                    if pg_procs.focus_prev(term.height // 3):
                        ui.start_interactive()
                elif keys.is_process_first(key):
                    if pg_procs.focus_first():
                        ui.start_interactive()
                elif keys.is_process_last(key):
                    if pg_procs.focus_last():
                        ui.start_interactive()
                elif key == keys.SPACE:
                    pg_procs.toggle_pin_focused()
                elif key == keys.COPY_TO_CLIPBOARD:
                    msg = pg_procs.copy_focused_query_to_clipboard()
                    msg_pile.send(msg)
                elif key.name == keys.CANCEL_SELECTION:
                    pg_procs.reset()
                    ui.end_interactive()
                elif keys.is_toggle_header_system(key):
                    ui.header.toggle_system()
                elif keys.is_toggle_header_instance(key):
                    ui.header.toggle_instance()
                elif keys.is_toggle_header_workers(key):
                    ui.header.toggle_workers()
                elif pg_procs.selected and key in (
                    keys.PROCESS_CANCEL,
                    keys.PROCESS_KILL,
                ):
                    action, color = {
                        keys.PROCESS_CANCEL: ("cancel", "yellow"),
                        keys.PROCESS_KILL: ("terminate", "red"),
                    }[key]
                    action_formatter = getattr(term, color)
                    pids = pg_procs.selected
                    if len(pids) > 1:
                        ptitle = f"processes {', '.join(str(p) for p in pids)}"
                    else:
                        ptitle = f"process {pids[0]}"
                    with term.location(x=0, y=term.height // 3):
                        print(
                            widgets.boxed(
                                term,
                                f"Confirm {action_formatter(action)} action on {ptitle}? (y/n)",
                                border_color=color,
                                center=True,
                                width=width,
                            ),
                            end="",
                        )
                        confirm_key = term.inkey(timeout=None)
                    if confirm_key.lower() == "y":
                        if action == "cancel":
                            for pid in pids:
                                data.pg_cancel_backend(pid)
                            msg_pile.send(
                                action_formatter(f"{ptitle.capitalize()} cancelled")
                            )
                        elif action == "terminate":
                            for pid in pids:
                                data.pg_terminate_backend(pid)
                            msg_pile.send(
                                action_formatter(f"{ptitle.capitalize()} terminated")
                            )
                        pg_procs.reset()
                        ui.end_interactive()
                        if wait_on_actions:
                            # Used in tests.
                            time.sleep(wait_on_actions)
                else:
                    pg_procs.reset()
                    ui.end_interactive()
                    changes: dict[str, Any] = {
                        "duration_mode": handlers.duration_mode(key, ui.duration_mode),
                        "wrap_query": handlers.wrap_query(key, ui.wrap_query),
                    }
                    if key in (keys.REFRESH_TIME_INCREASE, keys.REFRESH_TIME_DECREASE):
                        changes["refresh_time"] = handlers.refresh_time(
                            key, ui.refresh_time
                        )
                    query_mode = handlers.query_mode(key)
                    if query_mode is not None:
                        changes["query_mode"] = query_mode
                    else:
                        query_mode = ui.query_mode
                    sort_key = handlers.sort_key_for(key, query_mode, flag)
                    if sort_key is not None:
                        changes["sort_key"] = sort_key
                    ui.evolve(**changes)

            if in_help:
                # Only draw help screen once.
                if key is not None:
                    print(term.clear + term.home, end="")
                    views.help(
                        term,
                        __version__,
                        is_local,
                        lines_counter=views.line_counter(term.height),
                    )

            else:
                if not ui.in_pause and not ui.interactive():
                    if not options.dbsize and not skip_db_size:
                        skip_db_size = True

                    server_information = data.pg_get_server_information(
                        prev_server_info=server_information,
                        using_rds=options.rds,
                        skip_db_size=skip_db_size,
                        skip_tempfile=not options.tempfiles,
                        skip_walreceiver=not options.walreceiver,
                    )
                    memory, swap, load = activities.mem_swap_load()
                    system_info = types.SystemInfo.default(
                        memory=memory, swap=swap, load=load
                    )

                    if is_local:
                        memory, swap, load = activities.mem_swap_load()
                        system_info = attr.evolve(
                            system_info,
                            memory=memory,
                            swap=swap,
                            load=load,
                        )

                    if ui.query_mode == types.QueryMode.activities:
                        pg_procs.set_items(data.pg_get_activities(ui.duration_mode))
                        if is_local:
                            # TODO: Use this logic in waiting and blocking cases.
                            local_pg_procs, io_read, io_write = activities.ps_complete(
                                cast(list[types.RunningProcess], pg_procs.items),
                                sys_procs,
                                fs_blocksize,
                            )
                            system_info = attr.evolve(
                                system_info,
                                io_read=io_read,
                                io_write=io_write,
                                max_iops=activities.update_max_iops(
                                    system_info.max_iops,
                                    io_read.count,
                                    io_write.count,
                                ),
                            )
                            pg_procs.set_items(local_pg_procs)

                    else:
                        if ui.query_mode == types.QueryMode.blocking:
                            pg_procs.set_items(data.pg_get_blocking(ui.duration_mode))
                        elif ui.query_mode == types.QueryMode.waiting:
                            pg_procs.set_items(data.pg_get_waiting(ui.duration_mode))
                        else:
                            assert False  # help type checking

                    activity_stats = (pg_procs, system_info) if is_local else pg_procs  # type: ignore[assignment]

                if options.output is not None:
                    custom_asdict = partial(attr.asdict, recurse=True)
                    with open(options.output, "a") as f:
                        utils.csv_write(f, map(custom_asdict, pg_procs.items))

                views.screen(
                    term,
                    ui,
                    host=host,
                    pg_version=data.pg_version,
                    server_information=server_information,
                    activity_stats=activity_stats,
                    message=msg_pile.get(),
                    render_header=render_header,
                    render_footer=render_footer,
                    width=width,
                )

                if ui.interactive():
                    if not pg_procs.pinned:
                        ui.tick_interactive()
                elif pg_procs.selected:
                    pg_procs.reset()

            key = term.inkey(timeout=ui.refresh_time) or None



from __future__ import annotations

import logging
import os
from collections.abc import Callable, Sequence
from typing import Any, TypeVar, overload

Row = TypeVar("Row")

try:
    if "_PGACTIVITY_USE_PSYCOPG2" in os.environ:
        raise ImportError("psycopg2 requested from environment")

    import psycopg
    from psycopg import sql as sql
    from psycopg._encodings import pg2pyenc
    from psycopg.adapt import Buffer, Loader
    from psycopg.conninfo import conninfo_to_dict, make_conninfo
    from psycopg.errors import FeatureNotSupported as FeatureNotSupported
    from psycopg.errors import InsufficientPrivilege as InsufficientPrivilege
    from psycopg.errors import InterfaceError as InterfaceError
    from psycopg.errors import InvalidPassword as InvalidPassword
    from psycopg.errors import NotSupportedError
    from psycopg.errors import OperationalError as OperationalError
    from psycopg.errors import ProgrammingError as ProgrammingError
    from psycopg.errors import QueryCanceled as QueryCanceled
    from psycopg.rows import dict_row

    __version__ = psycopg.__version__

    Connection = psycopg.Connection[dict[str, Any]]

    class BytesLoader(Loader):
        def load(self, data: Buffer) -> bytes:
            if isinstance(data, memoryview):
                return bytes(data)
            return data

    class AutoTextLoader(Loader):
        def load(self, data: Buffer) -> str:
            if not isinstance(data, bytes):
                data = bytes(data)
            return data.decode(errors="replace")

    def connect(dsn: str = "", **kwargs: Any) -> Connection:
        if "PGCLIENTENCODING" not in os.environ:
            # Set client_encoding to 'auto', if not set by the user.
            # This is (more or less) what's done by psql.
            conninfo = conninfo_to_dict(dsn)
            conninfo.setdefault("client_encoding", "auto")
            dsn = make_conninfo("", **conninfo)
        conn = psycopg.connect(dsn, autocommit=True, row_factory=dict_row, **kwargs)
        if conn.info.encoding == "ascii":
            # If client encoding is still 'ascii', fall back to a loader with a replace
            # policy.
            logging.getLogger("pgactivity").warning(
                "client encoding is 'ascii', using a fallback loader for character types"
            )
            conn.adapters.register_loader("text", AutoTextLoader)
            conn.adapters.register_loader("varchar", AutoTextLoader)
            conn.adapters.register_loader('"char"', AutoTextLoader)
        return conn

    def server_version(conn: Connection) -> int:
        return conn.info.server_version

    def connection_parameters(conn: Connection) -> dict[str, Any]:
        return conn.info.get_parameters()

    def execute(
        conn: Connection,
        query: str | sql.Composed,
        args: None | Sequence[Any] | dict[str, Any] = None,
    ) -> None:
        conn.execute(query, args, prepare=True)

    @overload
    def cursor(
        conn: Connection, mkrow: Callable[..., Row], text_as_bytes: bool
    ) -> psycopg.Cursor[Row]: ...

    @overload
    def cursor(
        conn: Connection, mkrow: None, text_as_bytes: bool
    ) -> psycopg.Cursor[psycopg.rows.DictRow]: ...

    def cursor(
        conn: Connection, mkrow: Callable[..., Row] | None, text_as_bytes: bool
    ) -> psycopg.Cursor[psycopg.rows.DictRow] | psycopg.Cursor[Row]:
        if mkrow is not None:
            cur = conn.cursor(row_factory=psycopg.rows.kwargs_row(mkrow))
        else:
            cur = conn.cursor()  # type: ignore[assignment]
        if text_as_bytes:
            cur.adapters.register_loader("text", BytesLoader)
        return cur

    @overload
    def fetchone(
        conn: Connection,
        query: str | sql.Composed,
        args: None | Sequence[Any] | dict[str, Any] = None,
        *,
        mkrow: Callable[..., Row],
        text_as_bytes: bool = False,
    ) -> Row: ...

    @overload
    def fetchone(
        conn: Connection,
        query: str | sql.Composed,
        args: None | Sequence[Any] | dict[str, Any] = None,
        *,
        text_as_bytes: bool = False,
    ) -> dict[str, Any]: ...

    def fetchone(
        conn: Connection,
        query: str | sql.Composed,
        args: None | Sequence[Any] | dict[str, Any] = None,
        *,
        mkrow: Callable[..., Row] | None = None,
        text_as_bytes: bool = False,
    ) -> dict[str, Any] | Row:
        with cursor(conn, mkrow, text_as_bytes) as cur:
            row = cur.execute(query, args, prepare=True).fetchone()
        assert row is not None
        return row

    @overload
    def fetchall(
        conn: Connection,
        query: str | sql.Composed,
        args: None | Sequence[Any] | dict[str, Any] = None,
        *,
        mkrow: Callable[..., Row],
        text_as_bytes: bool = False,
    ) -> list[Row]: ...

    @overload
    def fetchall(
        conn: Connection,
        query: str | sql.Composed,
        args: None | Sequence[Any] | dict[str, Any] = None,
        *,
        text_as_bytes: bool = False,
    ) -> list[dict[str, Any]]: ...

    def fetchall(
        conn: Connection,
        query: str | sql.Composed,
        args: None | Sequence[Any] | dict[str, Any] = None,
        *,
        text_as_bytes: bool = False,
        mkrow: Callable[..., Row] | None = None,
    ) -> list[dict[str, Any]] | list[Row]:
        with cursor(conn, mkrow, text_as_bytes) as cur:
            return cur.execute(query, args, prepare=True).fetchall()

    def decode(value: bytes, pgenc: bytes, *, errors: str) -> str:
        """Decode 'value' with PostgreSQL encoding 'pgenc' converted to Python encoding
        name if available.
        """
        try:
            pyenc = pg2pyenc(pgenc)
        except NotSupportedError:
            pyenc = "utf-8"
        return value.decode(pyenc, errors=errors)

    def needs_password(exc: OperationalError) -> bool:
        assert exc.pgconn is not None
        return exc.pgconn.needs_password

except ImportError:
    import codecs

    import psycopg2
    import psycopg2.extensions
    from psycopg2 import sql as sql  # type: ignore[no-redef]

    # isort: off
    from psycopg2.errors import FeatureNotSupported as FeatureNotSupported  # type: ignore[assignment]
    from psycopg2.errors import InsufficientPrivilege as InsufficientPrivilege  # type: ignore[assignment]
    from psycopg2.errors import InterfaceError as InterfaceError  # type: ignore[assignment]
    from psycopg2.errors import InvalidPassword as InvalidPassword  # type: ignore[assignment]
    from psycopg2.errors import OperationalError as OperationalError  # type: ignore[assignment]
    from psycopg2.errors import ProgrammingError as ProgrammingError  # type: ignore[assignment]
    from psycopg2.errors import QueryCanceled as QueryCanceled  # type: ignore[assignment]

    # isort: on
    from psycopg2.extensions import connection as Connection  # type: ignore[assignment]
    from psycopg2.extras import DictCursor

    __version__ = psycopg2.__version__  # type: ignore[attr-defined]

    def connect(dsn: str = "", **kwargs: Any) -> Connection:
        try:
            kwargs.setdefault("database", kwargs.pop("dbname"))
        except KeyError:
            pass
        conn = psycopg2.connect(dsn, cursor_factory=DictCursor, **kwargs)
        conn.autocommit = True
        return conn  # type: ignore[no-any-return]

    def server_version(conn: Connection) -> int:
        return conn.server_version  # type: ignore[attr-defined, no-any-return]

    def connection_parameters(conn: Connection) -> dict[str, Any]:
        return conn.info.dsn_parameters  # type: ignore[attr-defined, no-any-return]

    def execute(
        conn: Connection,
        query: str | sql.Composed,
        args: None | Sequence[Any] | dict[str, Any] = None,
    ) -> None:
        with conn.cursor() as cur:
            cur.execute(query, args)

    def fetchone(  # type: ignore[no-redef]
        conn: Connection,
        query: str | sql.Composed,
        args: None | Sequence[Any] | dict[str, Any] = None,
        *,
        mkrow: Callable[..., Row] | None = None,
        text_as_bytes: bool = False,
    ) -> dict[str, Any] | Row:
        with conn.cursor() as cur:
            if text_as_bytes:
                psycopg2.extensions.register_type(psycopg2.extensions.BYTES, cur)  # type: ignore[arg-type]
            cur.execute(query, args)
            row = cur.fetchone()
        assert row is not None
        if mkrow is not None:
            return mkrow(**row)
        return row

    def fetchall(  # type: ignore[no-redef]
        conn: Connection,
        query: str | sql.Composed,
        args: None | Sequence[Any] | dict[str, Any] = None,
        *,
        mkrow: Callable[..., Row] | None = None,
        text_as_bytes: bool = False,
    ) -> list[dict[str, Any]] | list[Row]:
        with conn.cursor() as cur:
            if text_as_bytes:
                psycopg2.extensions.register_type(psycopg2.extensions.BYTES, cur)  # type: ignore[arg-type]
            cur.execute(query, args)
            rows = cur.fetchall()
        if mkrow is not None:
            return [mkrow(**row) for row in rows]
        return rows

    def decode(value: bytes, pgenc: bytes, *, errors: str) -> str:
        """Decode 'value' with PostgreSQL encoding 'pgenc' converted to Python encoding
        name if available.
        """
        try:
            pyenc = codecs.lookup(pgenc.decode()).name
        except LookupError:
            pyenc = "utf-8"
        return value.decode(pyenc, errors=errors)

    def needs_password(exc: OperationalError) -> bool:
        if isinstance(exc, InvalidPassword):
            return True
        msg = str(exc)
        return (
            msg.startswith("FATAL:  password authentication failed for user")
            or "fe_sendauth: no password supplied" in msg
        )


__all__ = [
    "__version__",
    "Connection",
    "FeatureNotSupported",
    "InterfaceError",
    "InvalidPassword",
    "InsufficientPrivilege",
    "OperationalError",
    "ProgrammingError",
    "QueryCanceled",
    "connect",
    "connection_parameters",
    "decode",
    "execute",
    "fetchall",
    "fetchone",
    "server_version",
    "sql",
]



from __future__ import annotations

import builtins
import os
import time
from collections.abc import Sequence
from typing import TypeVar
from warnings import catch_warnings, simplefilter

import attr
import psutil

from .types import (
    BlockingProcess,
    IOCounter,
    LoadAverage,
    LocalRunningProcess,
    MemoryInfo,
    RunningProcess,
    SortKey,
    SwapInfo,
    SystemProcess,
    WaitingProcess,
)


def sys_get_proc(pid: int) -> SystemProcess | None:
    """Return a SystemProcess instance matching given pid or None if access with psutil
    is not possible.
    """
    try:
        psproc = psutil.Process(pid)
        meminfo = psproc.memory_info()
        mem_percent = psproc.memory_percent()
        cpu_percent = psproc.cpu_percent(interval=0)
        cpu_times = psproc.cpu_times()
        io_counters = psproc.io_counters()
        status_iow = str(psproc.status())
    except (psutil.NoSuchProcess, psutil.AccessDenied):
        return None

    return SystemProcess(
        meminfo=meminfo,
        io_read=IOCounter(io_counters.read_count, io_counters.read_bytes),
        io_write=IOCounter(io_counters.write_count, io_counters.write_bytes),
        io_time=time.time(),
        mem_percent=mem_percent,
        cpu_percent=cpu_percent,
        cpu_times=cpu_times,
        read_delta=0,
        write_delta=0,
        io_wait=(status_iow == psutil.STATUS_DISK_SLEEP),
        psutil_proc=psproc,
    )


def ps_complete(
    pg_processes: Sequence[RunningProcess],
    processes: dict[int, SystemProcess],
    fs_blocksize: int,
) -> tuple[list[LocalRunningProcess], IOCounter, IOCounter]:
    """Transform the sequence of 'pg_processes' (RunningProcess) as LocalRunningProcess
    with local system information from the 'processes' map. Return LocalRunningProcess
    list, as well as read and write IO counters.

    The 'processes' map is updated in place.
    """
    local_procs = []
    read_bytes_delta = 0.0
    write_bytes_delta = 0.0
    read_count_delta = 0
    write_count_delta = 0
    n_io_time = time.time()
    for pg_proc in pg_processes:
        pid = pg_proc.pid
        new_proc = sys_get_proc(pid)
        if new_proc is None:
            continue
        try:
            # Getting information from the previous loop
            proc = processes[pid]
        except KeyError:
            # No previous information about this process
            proc = new_proc
        else:
            # Update old process with new information
            mem_percent = proc.mem_percent
            cpu_percent = proc.cpu_percent
            if proc.psutil_proc is not None:
                try:
                    mem_percent = proc.psutil_proc.memory_percent()
                    cpu_percent = proc.psutil_proc.cpu_percent(interval=0)
                except (psutil.NoSuchProcess, psutil.AccessDenied):
                    pass
            proc = attr.evolve(
                proc,
                io_wait=new_proc.io_wait,
                read_delta=(
                    (new_proc.io_read.bytes - proc.io_read.bytes)
                    / (n_io_time - proc.io_time)
                ),
                write_delta=(
                    (new_proc.io_write.bytes - proc.io_write.bytes)
                    / (n_io_time - proc.io_time)
                ),
                io_read=new_proc.io_read,
                io_write=new_proc.io_write,
                io_time=n_io_time,
                mem_percent=mem_percent,
                cpu_percent=cpu_percent,
            )

            # Global io counters
            read_bytes_delta += proc.read_delta
            write_bytes_delta += proc.write_delta

        processes[pid] = proc

        local_procs.append(
            LocalRunningProcess.from_process(
                pg_proc,
                cpu=proc.cpu_percent,
                mem=proc.mem_percent,
                read=proc.read_delta,
                write=proc.write_delta,
                io_wait=proc.io_wait,
            )
        )

    # store io counters
    if read_bytes_delta > 0:
        read_count_delta += int(read_bytes_delta / fs_blocksize)
    if write_bytes_delta > 0:
        write_count_delta += int(write_bytes_delta / fs_blocksize)

    io_read = IOCounter(count=read_count_delta, bytes=int(read_bytes_delta))
    io_write = IOCounter(count=write_count_delta, bytes=int(write_bytes_delta))

    return local_procs, io_read, io_write


T = TypeVar("T", RunningProcess, WaitingProcess, BlockingProcess, LocalRunningProcess)


def sorted(processes: list[T], *, key: SortKey, reverse: bool = False) -> list[T]:
    """Return processes sorted.

    >>> from ipaddress import IPv4Interface, ip_address

    PostgreSQL 13+
    >>> processes = [
    ...     LocalRunningProcess(
    ...         pid="6240",
    ...         xmin="1234",
    ...         application_name="pgbench",
    ...         database="pgbench",
    ...         user="postgres",
    ...         client=ip_address("127.0.0.2"),
    ...         cpu=0.1,
    ...         mem=0.993_254_939_413_836,
    ...         read=0.1,
    ...         write=0.282_725_318_098_656_75,
    ...         state="idle in transaction",
    ...         query="UPDATE pgbench_accounts SET abalance = abalance + 3062 WHERE aid = 1932841;",
    ...         encoding="UTF-8",
    ...         duration=0.1,
    ...         wait="ClientRead",
    ...         io_wait=False,
    ...         query_leader_pid=6240,
    ...         is_parallel_worker=False,
    ...     ),
    ...     LocalRunningProcess(
    ...         pid="6239",
    ...         xmin="2345",
    ...         application_name="pgbench",
    ...         database="pgbench",
    ...         user="postgres",
    ...         client=IPv4Interface("192.0.2.5/24"),
    ...         cpu=0.1,
    ...         mem=0.994_254_939_413_836,
    ...         read=0.1,
    ...         write=0.282_725_318_098_656_75,
    ...         state="idle in transaction",
    ...         query="UPDATE pgbench_accounts SET abalance = abalance + 141 WHERE aid = 7289374;",
    ...         encoding=None,
    ...         duration=0.1,
    ...         wait="ClientRead",
    ...         io_wait=False,
    ...         query_leader_pid=6239,
    ...         is_parallel_worker=False,
    ...     ),
    ...     LocalRunningProcess(
    ...         pid="6228",
    ...         xmin="3456",
    ...         application_name="pgbench",
    ...         database="pgbench",
    ...         user="postgres",
    ...         client=ip_address("2001:db8::"),
    ...         cpu=0.2,
    ...         mem=1.024_758_418_061_11,
    ...         read=0.2,
    ...         write=0.113_090_128_201_154_74,
    ...         state="active",
    ...         query="UPDATE pgbench_accounts SET abalance = abalance + 3062 WHERE aid = 1932841;",
    ...         encoding="UTF-8",
    ...         duration=0.1,
    ...         wait=False,
    ...         io_wait=False,
    ...         query_leader_pid=6240,
    ...         is_parallel_worker=True,
    ...     ),
    ... ]

    >>> processes = sorted(processes, key=SortKey.cpu, reverse=True)
    >>> [p.pid for p in processes]
    ['6228', '6240', '6239']
    >>> processes = sorted(processes, key=SortKey.mem)
    >>> [p.pid for p in processes]
    ['6240', '6239', '6228']

    When using the 'duration' sort key, processes are also sorted by ascending
    (query_leader_pid, is_parallel_worker).
    >>> processes = sorted(processes, key=SortKey.duration, reverse=True)
    >>> [p.pid for p in processes]
    ['6239', '6240', '6228']

    PostgreSQL 12- (query_leader_pid is None)
    >>> processes = [
    ...     LocalRunningProcess(
    ...         pid="6240",
    ...         xmin="1234",
    ...         application_name="pgbench",
    ...         database="pgbench",
    ...         user="postgres",
    ...         client=ip_address("192.168.1.2"),
    ...         cpu=0.1,
    ...         mem=0.993_254_939_413_836,
    ...         read=0.1,
    ...         write=0.282_725_318_098_656_75,
    ...         state="idle in transaction",
    ...         query="UPDATE pgbench_accounts SET abalance = abalance + 3062 WHERE aid = 1932841;",
    ...         encoding=None,
    ...         duration=0.1,
    ...         wait="ClientRead",
    ...         io_wait=False,
    ...         query_leader_pid=None,
    ...         is_parallel_worker=False,
    ...     ),
    ...     LocalRunningProcess(
    ...         pid="6239",
    ...         xmin="2345",
    ...         application_name="pgbench",
    ...         database="pgbench",
    ...         user="postgres",
    ...         client=ip_address("0000:0000:0000:0000:0000:0abc:0007:0def"),
    ...         cpu=0.1,
    ...         mem=0.994_254_939_413_836,
    ...         read=0.1,
    ...         write=0.282_725_318_098_656_75,
    ...         state="idle in transaction",
    ...         query="UPDATE pgbench_accounts SET abalance = abalance + 141 WHERE aid = 7289374;",
    ...         encoding="UTF-8",
    ...         duration=0.1,
    ...         wait="ClientRead",
    ...         io_wait=False,
    ...         query_leader_pid=None,
    ...         is_parallel_worker=False,
    ...     ),
    ...     LocalRunningProcess(
    ...         pid="6228",
    ...         xmin="3456",
    ...         application_name="pgbench",
    ...         database="pgbench",
    ...         user="postgres",
    ...         client=None,
    ...         cpu=0.2,
    ...         mem=1.024_758_418_061_11,
    ...         read=0.2,
    ...         write=0.113_090_128_201_154_74,
    ...         state="active",
    ...         query="UPDATE pgbench_accounts SET abalance = abalance + 3062 WHERE aid = 1932841;",
    ...         encoding="latin1",
    ...         duration=0.1,
    ...         wait=False,
    ...         io_wait=False,
    ...         query_leader_pid=None,
    ...         is_parallel_worker=True,
    ...     ),
    ... ]

    >>> processes = sorted(processes, key=SortKey.duration, reverse=True)
    >>> [p.pid for p in processes]
    ['6240', '6239', '6228']
    """

    # If we filter by duration, we also need to filter by ascending
    # (query_leader_pid, is_parallel_worker):
    # * for pg13+: query_leader_pid = coalesce(leader_pid, pid)
    # * for pg12-: query_leader_pid = Null / None
    # Note: parallel_worker have the same "duration" as their leader.
    if key == SortKey.duration:
        processes = builtins.sorted(
            processes,
            key=lambda p: (p.query_leader_pid, p.is_parallel_worker),
            reverse=False,
        )

    return builtins.sorted(
        processes,
        key=lambda p: getattr(p, key.name) or 0,  # TODO: avoid getattr()
        reverse=reverse,
    )


def update_max_iops(max_iops: int, read_count: float, write_count: float) -> int:
    """Update 'max_iops' value from read_count/write_count.

    >>> update_max_iops(45657, 123, 888)
    45657
    >>> update_max_iops(3, 123, 888)
    1011
    """
    return max(int(read_count + write_count), max_iops)


def get_load_average() -> tuple[float, float, float]:
    """Get load average"""
    return os.getloadavg()


def get_mem_swap() -> tuple[MemoryInfo, SwapInfo]:
    """Get memory and swap usage"""
    with catch_warnings():
        simplefilter("ignore", RuntimeWarning)
        phymem = psutil.virtual_memory()
        vmem = psutil.swap_memory()
    # 'buffers' and 'cached' attributes are not available on some systems (e.g. OSX)
    buffers = getattr(phymem, "buffers", 0)
    cached = getattr(phymem, "cached", 0)
    mem_used = phymem.total - (phymem.free + buffers + cached)
    return (
        MemoryInfo(mem_used, buffers + cached, phymem.free, phymem.total),
        SwapInfo(vmem.used, vmem.free, vmem.total),
    )


def mem_swap_load() -> tuple[MemoryInfo, SwapInfo, LoadAverage]:
    """Read memory, swap and load average from Data object."""
    memory, swap = get_mem_swap()
    load = LoadAverage(*get_load_average())
    return memory, swap, load



from __future__ import annotations

from blessed import Terminal


def boxed(
    term: Terminal,
    content: str,
    *,
    border: bool = True,
    border_color: str = "white",
    center: bool = False,
    width: int | None = None,
) -> str:
    if border:
        border_width = term.length(content) + 2
        border_formatter = getattr(term, border_color)
        lines = [
            border_formatter("┌" + "─" * border_width + "┐"),
            " ".join(
                [border_formatter("│") + term.normal, content, border_formatter("│")]
            ),
            border_formatter("└" + "─" * border_width + "┘") + term.normal,
        ]
    else:
        # border is disabled in UI tests.
        lines = ["", content, ""]
    if center:
        if width is None:
            width = term.width
        lines = [term.center(line, width=width) for line in lines]
    return "\n".join(lines)



from __future__ import annotations

import enum
import functools
from collections.abc import Callable, Iterable, Iterator, Mapping, MutableSet, Sequence
from datetime import timedelta
from ipaddress import IPv4Address, IPv6Address
from typing import Any, TypeVar, Union, overload

import attr
import psutil
from attr import validators

from . import colors, compat, pg, utils
from .config import Configuration, Flag, HeaderSection, UISection


class Pct(float):
    """Used to distinguish percentage from float when displaying the header"""


T = TypeVar("T")
E = TypeVar("E", bound=enum.IntEnum)


def enum_next(e: E) -> E:
    """Return an increment value of given enum.

    >>> class Seasons(enum.IntEnum):
    ...     winter = 1
    ...     spring = 2
    ...     summer = 3
    ...     autumn = 4

    >>> enum_next(Seasons.winter).name
    'spring'
    >>> enum_next(Seasons.spring).name
    'summer'
    >>> enum_next(Seasons.autumn).name
    'winter'
    """
    return e.__class__((e.value % max(e.__class__)) + 1)


@attr.s(auto_attribs=True, frozen=True, slots=True)
class Filters:
    dbname: str | None = None

    @classmethod
    def from_options(cls, filters: Sequence[str]) -> Filters:
        fields = compat.fields_dict(cls)
        attrs = {}
        for f in filters:
            try:
                fname, regex = f.split(":", 1)
            except ValueError:
                raise ValueError(f"malformatted filter value '{f}'")
            if not regex:
                raise ValueError(f"empty regex in filter '{f}'")
            if fname in attrs:
                raise ValueError(f"got multiple filters '{fname}'")
            if fname not in fields:
                raise ValueError(f"unknown filter '{fname}'")
            attrs[fname] = regex
        return cls(**attrs)


NO_FILTER = Filters()


class SortKey(enum.Enum):
    cpu = enum.auto()
    mem = enum.auto()
    read = enum.auto()
    write = enum.auto()
    duration = enum.auto()

    @classmethod
    def default(cls) -> SortKey:
        return cls.duration


@enum.unique
class QueryMode(enum.Enum):
    activities = "running queries"
    waiting = "waiting queries"
    blocking = "blocking queries"

    @classmethod
    def default(cls) -> QueryMode:
        return cls.activities


@enum.unique
class DurationMode(enum.IntEnum):
    query = 1
    transaction = 2
    backend = 3


def if_none(default: str) -> Callable[[Any], str]:
    def transform(value: Any) -> str:
        if value is None:
            return default
        return str(value)

    return transform


@attr.s(auto_attribs=True, frozen=True, slots=True)
class Column:
    """A column in stats table.

    >>> c = Column("pid", "PID", mandatory=True, sort_key=SortKey.cpu,
    ...            min_width=6, max_width=6,
    ...            transform=lambda v: str(v)[::-1])
    >>> c.title_render()
    'PID   '
    >>> c.title_color(SortKey.cpu)
    'cyan'
    >>> c.title_color(SortKey.duration)
    'green'
    >>> c.render('1234')
    '4321  '
    >>> c.render('12345678')
    '876543'

    >>> c = attr.evolve(c, justify="right", min_width=4, max_width=5)
    >>> c.title_render()
    ' PID'
    >>> c.render('7654321')
    '12345'
    >>> c.render('21')
    '  12'
    """

    key: str = attr.ib(repr=False)
    name: str
    mandatory: bool = False
    sort_key: SortKey | None = None
    min_width: int = attr.ib(default=0, repr=False)
    max_width: int | None = attr.ib(default=None, repr=False)
    justify: str = attr.ib(
        "left", validator=validators.in_(["left", "center", "right"])
    )
    transform: Callable[[Any], str] = attr.ib(default=if_none(""), repr=False)
    default_color: str | None = attr.ib(default=None, repr=False)
    value_color: Callable[[Any], str | None] | None = attr.ib(default=None, repr=False)
    # A callable returning a color name given a cell value or None to fall back to
    # "default_color"

    _justify: Callable[[str], str] = attr.ib(init=False)

    def __attrs_post_init__(self) -> None:
        if self.justify == "left":

            def _justify(value: str) -> str:
                return value.ljust(self.min_width)[: self.max_width]

        elif self.justify == "right":

            def _justify(value: str) -> str:
                return value.rjust(self.min_width)[: self.max_width]

        elif self.justify == "center":

            def _justify(value: str) -> str:
                return value.center(self.min_width)[: self.max_width]

        object.__setattr__(self, "_justify", _justify)

    def title_render(self) -> str:
        return self._justify(self.name)

    def title_color(self, sort_by: SortKey) -> str:
        if self.sort_key == sort_by:
            return "cyan"  # TODO: define a Color enum
        return "green"

    def render(self, value: Any) -> str:
        return self._justify(self.transform(value))

    def color(self, value: Any) -> str | None:
        if callable(self.value_color):
            color = self.value_color(value)
            if color is not None:
                return color
        return self.default_color


@attr.s(auto_attribs=True, slots=True)
class UIHeader:
    """Configuration for the header of the UI."""

    show_instance: bool = True
    show_system: bool = True
    show_workers: bool = True

    @classmethod
    def make(cls, config: HeaderSection | None, **options: bool | None) -> UIHeader:
        """Return a UIHeader built from configuration and command-line options.

        Command-line options take precedence over configuration.

        >>> config = HeaderSection(show_instance=False, show_system=False)
        >>> UIHeader.make(config, show_instance=True, show_workers=False)
        UIHeader(show_instance=True, show_system=False, show_workers=False)
        """
        values = {}
        if config is not None:
            values.update(attr.asdict(config))
        values.update({k: v for k, v in options.items() if v is not None})
        return cls(**values)

    def toggle_system(self) -> None:
        """Toggle the 'show_system' attribute.

        >>> h = UIHeader()
        >>> h.show_system
        True
        >>> h.toggle_system()
        >>> h.show_system
        False
        >>> h.toggle_system()
        >>> h.show_system
        True
        """
        self.show_system = not self.show_system

    def toggle_instance(self) -> None:
        """Toggle the 'show_instance' attribute.

        >>> h = UIHeader()
        >>> h.show_instance
        True
        >>> h.toggle_instance()
        >>> h.show_instance
        False
        >>> h.toggle_instance()
        >>> h.show_instance
        True
        """
        self.show_instance = not self.show_instance

    def toggle_workers(self) -> None:
        """Toggle the 'show_workers' attribute.

        >>> h = UIHeader()
        >>> h.show_workers
        True
        >>> h.toggle_workers()
        >>> h.show_workers
        False
        >>> h.toggle_workers()
        >>> h.show_workers
        True
        """
        self.show_workers = not self.show_workers


@attr.s(auto_attribs=True, slots=True)
class UI:
    """State of the UI."""

    header: UIHeader
    columns_by_querymode: Mapping[QueryMode, tuple[Column, ...]]
    min_duration: float = 0.0
    duration_mode: DurationMode = attr.ib(
        default=DurationMode.query, converter=DurationMode
    )
    wrap_query: bool = False
    sort_key: SortKey = attr.ib(default=SortKey.default(), converter=SortKey)
    query_mode: QueryMode = attr.ib(default=QueryMode.activities, converter=QueryMode)
    refresh_time: float | int = 2
    in_pause: bool = False
    interactive_timeout: int | None = None

    @classmethod
    def make(
        cls,
        header: UIHeader | None = None,
        config: Configuration | None = None,
        flag: Flag = Flag.all(),
        *,
        max_db_length: int = 16,
        filters: Filters = NO_FILTER,
        **kwargs: Any,
    ) -> UI:
        if header is None:
            header = UIHeader.make(config.header() if config else None)

        possible_columns: dict[str, Column] = {}
        config_values = (
            {k: v for k, v in config.items() if isinstance(v, UISection)}
            if config is not None
            else {}
        )

        def add_column(
            f: Flag | str, key: str, name: str | None = None, **kwargs: Any
        ) -> None:
            if isinstance(f, Flag):
                assert f.name is not None, f"invalid flag: {flag}"
                cfgkey = f.name
            else:
                cfgkey = f
            try:
                cfg = config_values.pop(cfgkey.lower())
            except KeyError:
                pass
            else:
                assert isinstance(cfg, UISection), cfg
                if cfg.width is not None:
                    kwargs["min_width"] = kwargs["max_width"] = cfg.width
                if cfg.color is not None:
                    if kwargs.get("value_color"):
                        assert config is not None
                        raise config.error(
                            f"the color for '{name}' column (option '{key}') cannot be configured"
                        )
                    kwargs["default_color"] = cfg.color
            assert key not in possible_columns, f"duplicated key {key}"
            if isinstance(f, Flag) and not f & flag:
                return
            if name is None:
                name = cfgkey
            possible_columns[key] = Column(key=key, name=name, **kwargs)

        add_column(
            Flag.APPNAME,
            key="application_name",
            name="APP",
            min_width=16,
            max_width=16,
            justify="right",
            default_color="bold_black",
        )
        add_column(
            Flag.CLIENT,
            key="client",
            min_width=16,
            max_width=16,
            justify="right",
            transform=if_none("local"),
            default_color="cyan",
        )
        add_column(
            Flag.CPU,
            key="cpu",
            name="CPU%",
            min_width=6,
            sort_key=SortKey.cpu,
        )
        add_column(
            Flag.DATABASE,
            key="database",
            name="DATABASE(*)" if filters.dbname else "DATABASE",
            min_width=max_db_length,
            transform=functools.lru_cache(
                lambda v: utils.ellipsis(v, width=16) if v else "",
            ),
            sort_key=None,
            default_color="bold_black",
        )
        add_column(
            Flag.IOWAIT,
            key="io_wait",
            name="IOW",
            min_width=4,
            transform=utils.yn,
            value_color=colors.wait,
        )
        add_column(
            Flag.MEM,
            key="mem",
            name="MEM%",
            min_width=4,
            sort_key=SortKey.mem,
            transform=lambda v: str(round(v, 1)),
        )
        add_column(
            Flag.MODE,
            key="mode",
            min_width=16,
            max_width=16,
            justify="right",
            value_color=colors.lock_mode,
        )
        add_column(
            Flag.PID,
            key="pid",
            min_width=6,
            default_color="cyan",
        )
        add_column(
            "query",
            key="query",
            name="Query",
            min_width=2,
        )
        add_column(
            Flag.READ,
            key="read",
            name="READ/s",
            min_width=8,
            sort_key=SortKey.read,
            transform=utils.naturalsize,
        )
        add_column(
            Flag.RELATION,
            key="relation",
            min_width=9,
            max_width=9,
            justify="right",
            default_color="cyan",
        )
        add_column(
            "state",
            key="state",
            min_width=17,
            justify="right",
            transform=utils.short_state,
            value_color=colors.short_state,
        )
        add_column(
            Flag.TIME,
            key="duration",
            name="TIME+",
            min_width=9,
            justify="right",
            sort_key=SortKey.duration,
            transform=lambda v: utils.format_duration(v)[0],
            value_color=lambda v: utils.format_duration(v)[1],
        )
        add_column(
            Flag.TYPE,
            key="type",
            min_width=16,
            max_width=16,
            justify="right",
        )
        add_column(
            Flag.USER,
            key="user",
            min_width=16,
            max_width=16,
            justify="right",
            default_color="bold_black",
        )
        add_column(
            Flag.WAIT,
            key="wait",
            name="Waiting",
            min_width=16,
            max_width=16,
            justify="right",
            transform=utils.wait_status,
            value_color=colors.wait,
        )
        add_column(
            Flag.WRITE,
            key="write",
            name="WRITE/s",
            min_width=8,
            sort_key=SortKey.write,
            transform=utils.naturalsize,
        )
        add_column(
            Flag.XMIN,
            key="xmin",
            min_width=8,
            default_color="cyan",
        )

        columns_key_by_querymode: Mapping[QueryMode, list[str]] = {
            QueryMode.activities: [
                "pid",
                "xmin",
                "database",
                "application_name",
                "user",
                "client",
                "cpu",
                "mem",
                "read",
                "write",
                "duration",
                "wait",
                "io_wait",
                "state",
                "query",
            ],
            QueryMode.waiting: [
                "pid",
                "database",
                "application_name",
                "user",
                "client",
                "relation",
                "type",
                "mode",
                "duration",
                "state",
                "query",
            ],
            QueryMode.blocking: [
                "pid",
                "database",
                "application_name",
                "user",
                "client",
                "relation",
                "type",
                "mode",
                "duration",
                "wait",
                "state",
                "query",
            ],
        }

        def make_columns_for(query_mode: QueryMode) -> Iterator[Column]:
            for key in columns_key_by_querymode[query_mode]:
                try:
                    yield possible_columns[key]
                except KeyError:
                    pass

        columns_by_querymode = {qm: tuple(make_columns_for(qm)) for qm in QueryMode}
        assert (
            not config_values
        ), f"unprocessed configuration entries: {', '.join(sorted(config_values))}"
        return cls(header=header, columns_by_querymode=columns_by_querymode, **kwargs)

    def interactive(self) -> bool:
        return self.interactive_timeout is not None

    def start_interactive(self) -> None:
        """Start interactive mode.

        >>> ui = UI.make()
        >>> ui.start_interactive()
        >>> ui.interactive_timeout
        3
        """
        self.interactive_timeout = 3

    def end_interactive(self) -> None:
        """End interactive mode.

        >>> ui = UI.make()
        >>> ui.start_interactive()
        >>> ui.interactive_timeout
        3
        >>> ui.end_interactive()
        >>> ui.interactive_timeout
        """
        self.interactive_timeout = None

    def tick_interactive(self) -> None:
        """Tick timeout of interactive mode.

        >>> ui = UI.make()
        >>> ui.tick_interactive()
        Traceback (most recent call last):
            ...
        RuntimeError: cannot tick interactive mode
        >>> ui.start_interactive()
        >>> ui.interactive_timeout
        3
        >>> ui.tick_interactive()
        >>> ui.interactive_timeout
        2
        >>> ui.tick_interactive()
        >>> ui.interactive_timeout
        1
        >>> ui.tick_interactive()
        >>> ui.interactive_timeout
        >>> ui.tick_interactive()
        Traceback (most recent call last):
            ...
        RuntimeError: cannot tick interactive mode
        """
        if self.interactive_timeout is None:
            raise RuntimeError("cannot tick interactive mode")
        assert self.interactive_timeout > 0, self.interactive_timeout
        self.interactive_timeout = (self.interactive_timeout - 1) or None

    def toggle_pause(self) -> None:
        """Toggle 'in_pause' attribute.

        >>> ui = UI.make()
        >>> ui.in_pause
        False
        >>> ui.toggle_pause()
        >>> ui.in_pause
        True
        >>> ui.toggle_pause()
        >>> ui.in_pause
        False
        """
        self.in_pause = not self.in_pause

    def evolve(self, **changes: Any) -> None:
        """Return a new UI with 'changes' applied.

        >>> ui = UI.make()
        >>> ui.query_mode.value
        'running queries'
        >>> ui.evolve(query_mode=QueryMode.blocking, sort_key=SortKey.write)
        >>> ui.query_mode.value
        'blocking queries'
        >>> ui.sort_key.name
        'write'
        """
        if self.in_pause:
            return
        forbidden = set(changes) - {
            "duration_mode",
            "wrap_query",
            "sort_key",
            "query_mode",
            "refresh_time",
        }
        assert not forbidden, forbidden
        fields = attr.fields(self.__class__)
        for field_name, value in changes.items():
            field = getattr(fields, field_name)
            if field.converter:
                value = field.converter(value)
            setattr(self, field_name, value)

    def column(self, key: str) -> Column:
        """Return the column matching 'key'.

        >>> ui = UI.make()
        >>> ui.column("cpu")  # doctest: +ELLIPSIS
        Column(name='CPU%', mandatory=False, sort_key=...)
        >>> ui.column("gloups")
        Traceback (most recent call last):
          ...
        ValueError: gloups
        """
        for column in self.columns_by_querymode[self.query_mode]:
            if column.key == key:
                return column
        else:
            raise ValueError(key)

    def columns(self) -> tuple[Column, ...]:
        """Return the tuple of Column for current mode.

        >>> flag = Flag.PID | Flag.DATABASE | Flag.APPNAME | Flag.RELATION
        >>> ui = UI.make(flag=flag)
        >>> [c.name for c in ui.columns()]
        ['PID', 'DATABASE', 'APP', 'state', 'Query']
        """
        return self.columns_by_querymode[self.query_mode]


@attr.s(auto_attribs=True, frozen=True, slots=True)
class Host:
    hostname: str
    user: str
    host: str
    port: int
    dbname: str


@attr.s(auto_attribs=True, slots=True, frozen=True)
class SwapInfo:
    used: int
    free: int
    total: int

    @classmethod
    def default(cls) -> SwapInfo:
        return cls(0, 0, 0)

    @property
    def pct_used(self) -> Pct | None:
        if self.total == 0:
            # account for the zero swap case (#318)
            return None
        return Pct(self.used * 100 / self.total)

    @property
    def pct_free(self) -> Pct | None:
        if self.total == 0:
            # account for the zero swap case (#318)
            return None
        return Pct(self.free * 100 / self.total)


@attr.s(auto_attribs=True, slots=True, frozen=True)
class MemoryInfo:
    used: int
    buff_cached: int
    free: int
    total: int

    @classmethod
    def default(cls) -> MemoryInfo:
        return cls(0, 0, 0, 0)

    @property
    def pct_used(self) -> Pct | None:
        if self.total == 0:
            return None
        return Pct(self.used * 100 / self.total)

    @property
    def pct_free(self) -> Pct | None:
        if self.total == 0:
            return None
        return Pct(self.free * 100 / self.total)

    @property
    def pct_bc(self) -> Pct | None:
        if self.total == 0:
            return None
        return Pct(self.buff_cached * 100 / self.total)


@attr.s(auto_attribs=True, slots=True)
class LoadAverage:
    avg1: float
    avg5: float
    avg15: float

    @classmethod
    def default(cls) -> LoadAverage:
        return cls(0.0, 0.0, 0.0)


@attr.s(auto_attribs=True, frozen=True, slots=True)
class IOCounter:
    count: int
    bytes: int

    @classmethod
    def default(cls) -> IOCounter:
        return cls(0, 0)


@attr.s(auto_attribs=True, frozen=True, slots=True)
class SystemInfo:
    memory: MemoryInfo
    swap: SwapInfo
    load: LoadAverage
    io_read: IOCounter
    io_write: IOCounter
    max_iops: int = 0

    @classmethod
    def default(
        cls,
        *,
        memory: MemoryInfo | None = None,
        swap: SwapInfo | None = None,
        load: LoadAverage | None = None,
    ) -> SystemInfo:
        """Zero-value builder.

        >>> SystemInfo.default()  # doctest: +NORMALIZE_WHITESPACE
        SystemInfo(memory=MemoryInfo(used=0, buff_cached=0, free=0, total=0),
                   swap=SwapInfo(used=0, free=0, total=0),
                   load=LoadAverage(avg1=0.0, avg5=0.0, avg15=0.0),
                   io_read=IOCounter(count=0, bytes=0),
                   io_write=IOCounter(count=0, bytes=0), max_iops=0)
        """
        return cls(
            memory or MemoryInfo.default(),
            swap or SwapInfo.default(),
            load or LoadAverage.default(),
            IOCounter.default(),
            IOCounter.default(),
            0,
        )


@attr.s(frozen=False, auto_attribs=True, slots=True)
class FailedQueriesInfo:
    temp_file_query_failed: bool = False
    wal_receivers_query_failed: bool = False


@attr.s(frozen=True, auto_attribs=True, slots=True)
class TempFileInfo:
    temp_files: int
    temp_bytes: int


@attr.s(frozen=True, auto_attribs=True, slots=True)
class ServerInformation:
    # Fetched from the database
    xact_count: int
    insert: int
    update: int
    delete: int
    tuples_returned: int
    total_size: int
    blks_read: int
    blks_hit: int
    xact_commit: int
    xact_rollback: int
    max_dbname_length: int
    uptime: timedelta
    epoch: int  # an epoch,  used for the calculation of the tps & size_evolution
    active_connections: int
    idle: int
    idle_in_transaction: int
    idle_in_transaction_aborted: int
    total: int
    waiting: int
    max_connections: int
    autovacuum_workers: int | None
    autovacuum_max_workers: int
    logical_replication_workers: int | None
    parallel_workers: int | None
    max_logical_replication_workers: int | None
    max_parallel_workers: int | None
    max_worker_processes: int | None
    max_wal_senders: int | None
    max_replication_slots: int | None
    wal_senders: int | None
    wal_receivers: int | None
    replication_slots: int | None
    temporary_file: TempFileInfo | None
    # Computed in data.pg_get_server_information()
    size_evolution: float
    tps: int
    insert_per_second: int
    update_per_second: int
    delete_per_second: int
    tuples_returned_per_second: int
    cache_hit_ratio_last_snap: Pct | None = attr.ib(
        converter=attr.converters.optional(Pct)
    )
    rollback_ratio_last_snap: Pct | None = attr.ib(
        converter=attr.converters.optional(Pct)
    )

    @property
    def worker_processes(self) -> int | None:
        if self.parallel_workers is None and self.logical_replication_workers is None:
            return None
        else:
            return (0 if self.parallel_workers is None else self.parallel_workers) + (
                0
                if self.logical_replication_workers is None
                else self.logical_replication_workers
            )


class LockType(enum.Enum):
    """Type of lockable object

    https://www.postgresql.org/docs/current/view-pg-locks.html
    """

    relation = enum.auto()
    extend = enum.auto()
    page = enum.auto()
    tuple = enum.auto()
    transactionid = enum.auto()
    virtualxid = enum.auto()
    object = enum.auto()
    userlock = enum.auto()
    advisory = enum.auto()

    def __str__(self) -> str:
        # Custom str(self) for transparent rendering in views.
        assert isinstance(self.name, str)
        # TODO ^ remove this assert for mypy > 0.991
        return self.name


def locktype(value: str) -> LockType:
    try:
        return LockType[value]
    except KeyError as exc:
        raise ValueError(f"invalid lock type {exc}") from None


@attr.s(auto_attribs=True, slots=True)
class BaseProcess:
    pid: int
    application_name: str
    database: str | None
    user: str
    client: None | IPv4Address | IPv6Address
    duration: float | None
    state: str
    query: str | None
    encoding: str | None
    query_leader_pid: int | None
    is_parallel_worker: bool

    _P = TypeVar("_P", bound="BaseProcess")

    @classmethod
    def from_bytes(
        cls: type[_P],
        server_encoding: bytes,
        *,
        encoding: str | bytes | None,
        **kwargs: Any,
    ) -> _P:
        if encoding is None:
            enc = server_encoding
        elif isinstance(encoding, bytes):  # psycopg2
            enc, encoding = encoding, encoding.decode()
        elif isinstance(encoding, str):
            enc = encoding.encode()
        for name, value in kwargs.items():
            if isinstance(value, bytes):
                kwargs[name] = pg.decode(value, enc, errors="replace")
        return cls(encoding=encoding, **kwargs)


@attr.s(auto_attribs=True, frozen=True, slots=True)
class RunningProcess(BaseProcess):
    """Process for a running query."""

    xmin: int
    wait: bool | None | str
    query_leader_pid: int | None
    is_parallel_worker: bool


@attr.s(auto_attribs=True, frozen=True, slots=True)
class WaitingProcess(BaseProcess):
    """Process for a waiting query."""

    # Lock information from pg_locks view
    # https://www.postgresql.org/docs/current/view-pg-locks.html
    mode: str
    type: LockType = attr.ib(converter=locktype)
    relation: str

    # TODO: update queries to select/compute these column.
    query_leader_pid: int | None = attr.ib(default=None, init=False)
    is_parallel_worker: bool = attr.ib(default=False, init=False)


@attr.s(auto_attribs=True, frozen=True, slots=True)
class BlockingProcess(BaseProcess):
    """Process for a blocking query."""

    # Lock information from pg_locks view
    # https://www.postgresql.org/docs/current/view-pg-locks.html
    mode: str
    type: LockType = attr.ib(converter=locktype)
    relation: str
    wait: bool | None | str

    # TODO: update queries to select/compute these column.
    query_leader_pid: int | None = attr.ib(default=None, init=False)
    is_parallel_worker: bool = attr.ib(default=False, init=False)


@attr.s(auto_attribs=True, frozen=True, slots=True)
class SystemProcess:
    meminfo: tuple[int, ...]
    io_read: IOCounter
    io_write: IOCounter
    io_time: float
    mem_percent: float
    cpu_percent: float
    cpu_times: tuple[float, ...]
    read_delta: float
    write_delta: float
    io_wait: bool
    psutil_proc: psutil.Process | None


@attr.s(auto_attribs=True, frozen=True, slots=True)
class LocalRunningProcess(RunningProcess):
    cpu: float
    mem: float
    read: float
    write: float
    io_wait: bool

    @classmethod
    def from_process(
        cls, process: RunningProcess, **kwargs: float | str
    ) -> LocalRunningProcess:
        return cls(**dict(attr.asdict(process), **kwargs))


@attr.s(auto_attribs=True, slots=True)
class SelectableProcesses:
    """Selectable list of processes.

    >>> @attr.s(auto_attribs=True)
    ... class Proc:
    ...     pid: int

    >>> w = SelectableProcesses(list(map(Proc, [456, 123, 789])))
    >>> len(w)
    3

    Nothing focused at initialization:
    >>> w.focused

    >>> w.focus_next()
    True
    >>> w.focused
    456
    >>> w.focus_next()
    True
    >>> w.focused
    123
    >>> _ = w.focus_next(10)
    >>> w.focused
    789
    >>> _ = w.focus_next()
    >>> w.focused
    456
    >>> _ = w.focus_next()
    >>> w.focused
    123
    >>> w.focus_prev()
    True
    >>> w.focused
    456
    >>> w.focus_prev()
    True
    >>> w.focused
    789
    >>> w.focused = 789
    >>> w.focus_next(2)
    True
    >>> w.focused
    789
    >>> w.focus_prev()
    True
    >>> w.focused
    123
    >>> _ = w.focus_prev(3)
    >>> w.focused
    456
    >>> _ = w.focus_last()
    >>> w.focused
    789
    >>> _ = w.focus_first()
    >>> w.focused
    456
    >>> w.focused = 789
    >>> _ = w.focus_prev(2)
    >>> w.focused
    456
    >>> w.focused = 789
    >>> w.set_items(sorted(w.items))
    >>> w.focused
    789
    >>> w.focus_prev()
    True
    >>> w.focused
    456
    >>> w.focus_next()
    True
    >>> w.focused
    789

    >>> w.selected, w.focused
    ([789], 789)
    >>> w.toggle_pin_focused()
    >>> w.focus_next()
    True
    >>> w.toggle_pin_focused()
    >>> w.selected, w.focused
    ([123, 789], 123)
    >>> w.toggle_pin_focused()
    >>> w.focus_next()
    True
    >>> w.toggle_pin_focused()
    >>> w.selected, w.focused
    ([456, 789], 456)

    >>> w[1]
    Proc(pid=456)
    >>> w[1:3]
    [Proc(pid=456), Proc(pid=789)]

    >>> w.reset()
    >>> w.selected, w.focused
    ([], None)

    """

    items: list[BaseProcess]
    focused: int | None = None
    pinned: MutableSet[int] = attr.ib(default=attr.Factory(set))

    def __len__(self) -> int:
        return len(self.items)

    def __iter__(self) -> Iterator[BaseProcess]:
        return iter(self.items)

    @overload
    def __getitem__(self, i: int) -> BaseProcess: ...

    @overload
    def __getitem__(self, s: slice) -> list[BaseProcess]: ...

    def __getitem__(self, val: int | slice) -> BaseProcess | list[BaseProcess]:
        return self.items[val]

    @property
    def selected(self) -> list[int]:
        if self.pinned:
            return list(self.pinned)
        elif self.focused:
            return [self.focused]
        else:
            return []

    def reset(self) -> None:
        self.focused = None
        self.pinned.clear()

    def set_items(self, new_items: Sequence[BaseProcess]) -> None:
        self.items[:] = list(new_items)

    def position(self) -> int | None:
        if self.focused is None:
            return None
        for idx, proc in enumerate(self.items):
            if proc.pid == self.focused:
                return idx
        return None

    def focus_next(self, offset: int = 1) -> bool:
        if not self.items:
            return False
        idx = self.position()
        bottom = len(self.items) - 1
        if idx is None or (offset == 1 and idx == bottom):
            next_idx = 0
        else:
            next_idx = min(bottom, idx + offset)
        self.focused = self.items[next_idx].pid
        return True

    def focus_prev(self, offset: int = 1) -> bool:
        if not self.items:
            return False
        idx = self.position() or 0
        if offset == 1:
            next_idx = idx - offset
        else:
            next_idx = max(idx - offset, 0)
        self.focused = self.items[next_idx].pid
        return True

    def focus_first(self) -> bool:
        if not self.items:
            return False
        self.focused = self.items[0].pid
        return True

    def focus_last(self) -> bool:
        if not self.items:
            return False
        self.focused = self.items[-1].pid
        return True

    def toggle_pin_focused(self) -> None:
        assert self.focused is not None
        try:
            self.pinned.remove(self.focused)
        except KeyError:
            self.pinned.add(self.focused)

    def copy_focused_query_to_clipboard(self) -> str:
        """Copy focused query to system clipboard using ANSI OSC 52 escape sequence."""
        assert self.focused is not None
        for proc in self.items:
            if proc.pid == self.focused:
                break
        else:
            return "no focused process found"
        if proc.query is None:
            return "process has no query"

        utils.osc52_copy(proc.query)
        return f"query of process {proc.pid} copied to clipboard"


ActivityStats = Union[
    Iterable[WaitingProcess],
    Iterable[RunningProcess],
    tuple[Iterable[WaitingProcess], SystemInfo],
    tuple[Iterable[BlockingProcess], SystemInfo],
    tuple[Iterable[LocalRunningProcess], SystemInfo],
]



from __future__ import annotations

import functools
import inspect
import itertools
from collections.abc import Callable, Iterable, Iterator, Sequence
from textwrap import TextWrapper, dedent
from typing import Any, Literal

from blessed import Terminal

from . import colors, utils
from .activities import sorted as sorted_processes
from .compat import link
from .keys import (
    BINDINGS,
    EXIT_KEY,
)
from .keys import HELP as HELP_KEY
from .keys import (
    KEYS_BY_QUERYMODE,
    MODES,
    PAUSE_KEY,
    PROCESS_CANCEL,
    PROCESS_KILL,
    PROCESS_PIN,
    Key,
)
from .types import (
    UI,
    ActivityStats,
    Column,
    Host,
    IOCounter,
    Pct,
    SelectableProcesses,
    ServerInformation,
    SystemInfo,
)


class line_counter:
    def __init__(self, start: int) -> None:
        self.value = start

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.value})"

    def __next__(self) -> int:
        current_value = self.value
        self.value -= 1
        return current_value


@functools.lru_cache(maxsize=512)
def shorten(term: Terminal, text: str, width: int | None = None) -> str:
    r"""Truncate 'text' to fit in the given 'width' (or term.width).

    This is similar to textwrap.shorten() but sequence-aware.

    >>> term = Terminal()
    >>> text = f"{term.green('hello')}, world"
    >>> text
    'hello, world'
    >>> shorten(term, text, 6)
    'hello,'
    >>> shorten(term, text, 3)
    'hel'
    >>> shorten(term, "", 3)
    ''
    """
    if not text:
        return ""
    wrapped = term.wrap(text, width=width, max_lines=1)
    return wrapped[0] + term.normal


def limit(func: Callable[..., Iterable[str]]) -> Callable[..., None]:
    """View decorator handling screen height limit.

    >>> term = Terminal()

    >>> def view(term, n, *, prefix="line"):
    ...     for i in range(n):
    ...         yield f"{prefix} #{i}"

    >>> count = line_counter(2)
    >>> limit(view)(term, 3, lines_counter=count)
    line #0
    line #1
    >>> count
    line_counter(0)
    >>> count = line_counter(3)
    >>> limit(view)(term, 2, lines_counter=count)
    line #0
    line #1
    >>> count
    line_counter(1)
    >>> limit(view)(term, 3, prefix="row")
    row #0
    row #1
    row #2

    A single line is displayed with an EOL as well:
    >>> count = line_counter(10)
    >>> limit(view)(term, 1, lines_counter=count) or print("<--", end="")
    line #0
    <--
    >>> count
    line_counter(9)
    """

    @functools.wraps(func)
    def wrapper(term: Terminal, *args: Any, **kwargs: Any) -> None:
        counter = kwargs.pop("lines_counter", None)
        width = kwargs.pop("width", None)
        signature = inspect.signature(func)
        if "width" in signature.parameters:
            kwargs["width"] = width
        for line in func(term, *args, **kwargs):
            print(shorten(term, line, width) + term.clear_eol)
            if counter is not None and next(counter) == 1:
                break

    return wrapper


@limit
def help(term: Terminal, version: str, is_local: bool) -> Iterable[str]:
    """Render help menu."""
    project_url = "https://github.com/dalibo/pg_activity"
    intro = dedent(
        f"""\
    {term.bold_green}pg_activity {version} - {link(term, project_url, project_url)}
    {term.normal}Released under PostgreSQL License.
    """
    )

    def key_mappings(keys: Iterable[Key]) -> Iterable[str]:
        for key in keys:
            key_name = key.name or key.value
            yield f"{term.bright_cyan}{key_name.rjust(10)}{term.normal}: {key.description}"

    footer = "Press any key to exit."
    yield from intro.splitlines()
    yield ""

    bindings = BINDINGS
    if not is_local:
        bindings = [b for b in bindings if not b.local_only]
    yield from key_mappings(bindings)
    yield "Mode"
    yield from key_mappings(MODES)
    yield ""
    yield footer


@limit
def header(
    term: Terminal,
    ui: UI,
    *,
    host: Host,
    pg_version: str,
    server_information: ServerInformation,
    system_info: SystemInfo | None = None,
) -> Iterator[str]:
    @functools.singledispatch
    def render(x: Any) -> str:
        if x is None:
            return "-"
        raise AssertionError(f"not implemented for type '{type(x).__name__}'")

    @render.register(str)
    def render_str(s: str) -> str:
        return term.bold_green(s)

    @render.register(int)
    def render_int(n: int) -> str:
        return term.bold_green(str(n))

    @render.register(Pct)
    def render_pct(n: Pct) -> str:
        return term.bold_green(f"{n:.2f}%")

    @render.register(float)
    def render_float(n: float) -> str:
        return term.bold_green(f"{n:.2f}")

    @render.register(IOCounter)
    def render_iocounter(i: IOCounter) -> str:
        hbytes = utils.naturalsize(i.bytes) + "/s"
        counts = str(i.count) + "/s"
        return f"{term.bold_green(hbytes)} - {term.bold_green(counts)}"

    def render_columns(
        columns: Sequence[list[str]], *, delimiter: str = f"{term.blue(',')} "
    ) -> Iterator[str]:
        column_widths = [
            max(len(column_row) for column_row in column) for column in columns
        ]

        def indent(text: str) -> str:
            return " " + text

        for row in itertools.zip_longest(*columns, fillvalue=""):
            yield indent(
                "".join(
                    (cell + delimiter).ljust(width + len(delimiter))
                    for width, cell in zip(column_widths, row)
                )
            ).rstrip().rstrip(delimiter.strip())

    si = server_information

    """Return window header lines."""
    pg_host = f"{host.user}@{host.host}:{host.port}/{host.dbname}"
    yield (
        " - ".join(
            [
                pg_version,
                f"{term.bold}{host.hostname}{term.normal}",
                f"{term.cyan}{pg_host}{term.normal}",
                f"Ref.: {term.yellow}{ui.refresh_time}s{term.normal}",
                f"Duration mode: {term.yellow}{ui.duration_mode.name}{term.normal}",
            ]
            + (
                [f"Min. duration: {term.yellow}{ui.min_duration}s{term.normal}"]
                if ui.min_duration
                else []
            )
        )
    )

    total_size = utils.naturalsize(si.total_size)
    size_ev = f"{utils.naturalsize(si.size_evolution)}/s"
    uptime = utils.naturaltimedelta(si.uptime)

    if ui.header.show_instance:
        # First rows are always displayed, as the underlying data is always available.
        columns = [
            [f"* Global: {render(uptime)} uptime"],
            [f"{render(total_size)} dbs size - {render(size_ev)} growth"],
            [f"{render(si.cache_hit_ratio_last_snap)} cache hit ratio"],
            [f"{render(si.rollback_ratio_last_snap)} rollback ratio"],
        ]
        yield from render_columns(columns)

        columns = [
            [f"  Sessions: {render(si.total)}/{render(si.max_connections)} total"],
            [f"{render(si.active_connections)} active"],
            [f"{render(si.idle)} idle"],
            [f"{render(si.idle_in_transaction)} idle in txn"],
            [f"{render(si.idle_in_transaction_aborted)} idle in txn abrt"],
            [f"{render(si.waiting)} waiting"],
        ]
        yield from render_columns(columns)

        if si.temporary_file is not None:
            temp_files = si.temporary_file.temp_files
            temp_size = utils.naturalsize(si.temporary_file.temp_bytes)
        else:
            temp_files = None
            temp_size = None
        columns = [
            [f"  Activity: {render(si.tps)} tps"],
            [f"{render(si.insert_per_second)} insert/s"],
            [f"{render(si.update_per_second)} update/s"],
            [f"{render(si.delete_per_second)} delete/s"],
            [f"{render(si.tuples_returned_per_second)} tuples returned/s"],
            [f"{render(temp_files)} temp files"],
            [f"{render(temp_size)} temp size"],
        ]
        yield from render_columns(columns)
    if ui.header.show_workers:
        columns = [
            [
                f"* Worker processes: {render(si.worker_processes)}/{render(si.max_worker_processes)} total"
            ],
            [
                f"{render(si.logical_replication_workers)}/{render(si.max_logical_replication_workers)} logical workers"
            ],
            [
                f"{render(si.parallel_workers)}/{render(si.max_parallel_workers)} parallel workers"
            ],
        ]
        yield from render_columns(columns)

        columns = [
            [
                f"  Other processes & info: {render(si.autovacuum_workers)}/{render(si.autovacuum_max_workers)} autovacuum workers"
            ],
            [f"{render(si.wal_senders)}/{render(si.max_wal_senders)} wal senders"],
            [f"{render(si.wal_receivers)} wal receivers"],
            [
                f"{render(si.replication_slots)}/{render(si.max_replication_slots)} repl. slots"
            ],
        ]
        yield from render_columns(columns)

    # System information, only available in "local" mode.
    if system_info is not None and ui.header.show_system:
        used, bc, free, total = (
            utils.naturalsize(system_info.memory.used),
            utils.naturalsize(system_info.memory.buff_cached),
            utils.naturalsize(system_info.memory.free),
            utils.naturalsize(system_info.memory.total),
        )
        system_columns = [
            [f"* Mem.: {render(total)} total"],
            [f"{render(free)} ({render(system_info.memory.pct_free)}) free"],
            [f"{render(used)} ({render(system_info.memory.pct_used)}) used"],
            [f"{render(bc)} ({render(system_info.memory.pct_bc)}) buff+cached"],
        ]
        yield from render_columns(system_columns)

        used, free, total = (
            utils.naturalsize(system_info.swap.used),
            utils.naturalsize(system_info.swap.free),
            utils.naturalsize(system_info.swap.total),
        )
        system_columns = [
            [f"  Swap: {render(total)} total"],
            [f"{render(free)} ({render(system_info.swap.pct_free)}) free"],
            [f"{render(used)} ({render(system_info.swap.pct_used)}) used"],
        ]
        yield from render_columns(system_columns)

        iops = f"{system_info.max_iops}/s"
        system_columns = [
            [f"  IO: {render(iops)} max iops"],
            [f"{render(system_info.io_read)} read"],
            [f"{render(system_info.io_write)} write"],
        ]
        yield from render_columns(system_columns)

        load = system_info.load
        system_columns = [
            [
                f"  Load average: {render(load.avg1)} {render(load.avg5)} {render(load.avg15)}"
            ],
        ]
        yield from render_columns(system_columns)


@limit
def query_mode(term: Terminal, ui: UI) -> Iterator[str]:
    r"""Display query mode title.

    >>> from pgactivity.types import QueryMode, UI

    >>> term = Terminal()
    >>> ui = UI.make(query_mode=QueryMode.blocking)
    >>> query_mode(term, ui)
                                    BLOCKING QUERIES
    >>> ui = UI.make(query_mode=QueryMode.activities, in_pause=True)
    >>> query_mode(term, ui)  # doctest: +NORMALIZE_WHITESPACE
                                    PAUSE
    """
    if ui.in_pause:
        yield term.black_on_yellow(term.center("PAUSE", fillchar=" "))
    else:
        yield term.green_bold(
            term.center(ui.query_mode.value.upper(), fillchar=" ").rstrip()
        )


@limit
def columns_header(term: Terminal, ui: UI) -> Iterator[str]:
    """Yield columns header lines."""
    htitles = []
    for column in ui.columns():
        color = getattr(term, f"black_on_{column.title_color(ui.sort_key)}")
        htitles.append(f"{color}{column.title_render()}")
    yield term.ljust(" ".join(htitles), fillchar=" ") + term.normal


def get_indent(ui: UI) -> str:
    """Return indentation for Query column.

    >>> from pgactivity.config import Flag
    >>> from pgactivity.types import UI
    >>> ui = UI.make(flag=Flag.CPU)
    >>> get_indent(ui)
    '                         '
    >>> ui = UI.make(flag=Flag.PID | Flag.DATABASE | Flag.APPNAME | Flag.RELATION)
    >>> get_indent(ui)
    '                                                           '
    """
    return " " * sum(c.min_width + 1 for c in ui.columns() if c.name != "Query")


def format_query(query: str, is_parallel_worker: bool) -> str:
    r"""Return the query string formatted.

    >>> print(format_query("SELECT 1", True))
    \_ SELECT 1
    >>> format_query("SELECT   1", False)
    'SELECT 1'
    """
    prefix = r"\_ " if is_parallel_worker else ""
    return prefix + utils.clean_str(query)


@limit
def processes_rows(
    term: Terminal,
    ui: UI,
    processes: SelectableProcesses,
    maxlines: int,
    width: int | None,
) -> Iterator[str]:
    """Display table rows with processes information."""

    if width is None:
        width = term.width

    def cell(
        value: Any,
        column: Column,
        cursor: Literal["pinned", "focused"] | None,
    ) -> None:
        if cursor is not None:
            if cursor == "pinned":
                color_name = colors.PINNED_COLOR
            elif cursor == "focused":
                color_name = colors.FOCUSED_COLOR
            else:
                raise AssertionError(cursor)
        else:
            color_name = column.color(value) or "normal"
        color = getattr(term, color_name or "normal")
        # We also restore 'normal' style so that the next item does not
        # inherit from that of the previous one.
        text.append(f"{color}{column.render(value)}{term.normal}")

    position = processes.position()
    if position is None:
        display_processes = iter(processes)
    else:
        # Scrolling is handled here. We just have to manage the start position of the
        # display. the length is managed by @limit.
        if ui.wrap_query:
            # When the query is wrapped, we always display selected process first (sort
            # of relative scrolling).
            start = position
        else:
            # Otherwise, we compute the start position of the table and try to have a 5 lines
            # leeway at the bottom of the page to increase readability.
            start = 0
            bottom = int(maxlines // 5)
            if position is not None and position >= maxlines - bottom:
                start = position - maxlines + 1 + bottom

        display_processes = itertools.chain(
            iter(processes[-(len(processes) - start) :]), iter(processes[:start])
        )

    focused, pinned = processes.focused, processes.pinned

    for process in display_processes:
        cursor: Literal["focused", "pinned"] | None = None
        if process.pid == focused:
            cursor = "focused"
        elif process.pid in pinned:
            cursor = "pinned"
        text: list[str] = []
        for column in ui.columns():
            field = column.key
            if field != "query":
                cell(getattr(process, field), column, cursor=cursor)

        indent = get_indent(ui) + " "
        qwidth = width - len(indent)

        if qwidth > 0 and process.query is not None:
            query = format_query(process.query, process.is_parallel_worker)

            if not ui.wrap_query:
                query_value = query[:qwidth]
            else:
                wrapped_lines = TextWrapper(qwidth).wrap(query)
                query_value = f"\n{indent}".join(wrapped_lines)

            cell(query_value, ui.column("query"), cursor=cursor)

        yield from (" ".join(text) + term.normal).splitlines()


def footer_message(term: Terminal, message: str, width: int | None = None) -> None:
    if width is None:
        width = term.width
    print(term.center(message[:width]) + term.normal, end="")


def footer_help(term: Terminal, width: int | None = None) -> None:
    """Footer line with help keys."""
    query_modes_help = [
        ("/".join(keys[:-1]), qm.value) for qm, keys in KEYS_BY_QUERYMODE.items()
    ]
    assert PAUSE_KEY.name is not None
    footer_values = query_modes_help + [
        (PAUSE_KEY.name, PAUSE_KEY.description),
        (EXIT_KEY.value, EXIT_KEY.description),
        (HELP_KEY, "help"),
    ]
    render_footer(term, footer_values, width)


def render_footer(
    term: Terminal, footer_values: list[tuple[str, str]], width: int | None
) -> None:
    if width is None:
        width = term.width
    ncols = len(footer_values)
    column_width = (width - ncols - 1) // ncols

    def render_column(key: str, desc: str) -> str:
        col_width = column_width - term.length(key) - 1
        if col_width <= 0:
            return ""
        desc = term.ljust(desc[:col_width], width=col_width, fillchar=" ")
        return f"{key} {term.cyan_reverse(desc)}"

    row = " ".join(
        [render_column(key, desc.capitalize()) for key, desc in footer_values]
    )
    assert term.length(row) <= width, (term.length(row), width, ncols)
    print(term.ljust(row, width=width, fillchar=term.cyan_reverse(" ")), end="")


def footer_interative_help(term: Terminal, width: int | None = None) -> None:
    """Footer line with help keys for interactive mode."""
    assert PROCESS_PIN.name is not None
    footer_values = [
        (PROCESS_CANCEL, "cancel current query"),
        (PROCESS_KILL, "terminate underlying session"),
        (PROCESS_PIN.name, PROCESS_PIN.description),
        ("Other", "back to activities"),
        (EXIT_KEY.value, EXIT_KEY.description),
    ]
    return render_footer(term, footer_values, width)


def screen(
    term: Terminal,
    ui: UI,
    *,
    host: Host,
    pg_version: str,
    server_information: ServerInformation,
    activity_stats: ActivityStats,
    message: str | None,
    render_header: bool = True,
    render_footer: bool = True,
    width: int | None = None,
) -> None:
    """Display the screen."""

    system_info: SystemInfo | None
    if isinstance(activity_stats, tuple):
        processes, system_info = activity_stats
    else:
        processes, system_info = activity_stats, None
    processes.set_items(sorted_processes(processes, key=ui.sort_key, reverse=True))

    print(term.home, end="")
    top_height = term.height - (1 if render_footer else 0)
    lines_counter = line_counter(top_height)

    if render_header:
        header(
            term,
            ui,
            host=host,
            pg_version=pg_version,
            server_information=server_information,
            system_info=system_info,
            lines_counter=lines_counter,
            width=width,
        )

    query_mode(term, ui, lines_counter=lines_counter, width=width)
    columns_header(term, ui, lines_counter=lines_counter, width=width)
    processes_rows(
        term,
        ui,
        processes,
        maxlines=lines_counter.value,  # Used by process_rows
        lines_counter=lines_counter,  # Used by @limit
        width=width,
    )

    # Clear remaining lines in screen until footer (or EOS)
    print(f"{term.clear_eol}\n" * lines_counter.value, end="")

    if render_footer:
        with term.location(x=0, y=top_height):
            if message is not None:
                footer_message(term, message, width)
            elif ui.interactive():
                footer_interative_help(term, width)
            else:
                footer_help(term, width)



from __future__ import annotations

import getpass
import logging
import re
from argparse import Namespace
from functools import partial

import attr
import psutil

from . import pg, queries
from .pg import Connection, sql
from .types import (
    NO_FILTER,
    BlockingProcess,
    FailedQueriesInfo,
    Filters,
    Pct,
    RunningProcess,
    ServerInformation,
    TempFileInfo,
    WaitingProcess,
)

logger = logging.getLogger("pgactivity")


def pg_get_version(pg_conn: Connection) -> str:
    """Get PostgreSQL server version."""
    ret = pg.fetchone(pg_conn, queries.get("get_version"))
    return ret["pg_version"]  # type: ignore[no-any-return]


def pg_get_short_version(text_version: str) -> str:
    """Return PostgreSQL short version from a string (SELECT version()).

    >>> pg_get_short_version('PostgreSQL 11.9')
    'PostgreSQL 11.9'
    >>> pg_get_short_version('EnterpriseDB 11.9 (Debian 11.9-0+deb10u1)')
    'EnterpriseDB 11.9'
    >>> pg_get_short_version("PostgreSQL 9.3.24 on x86_64-pc-linux-gnu (Debian 9.3.24-1.pgdg80+1), compiled by gcc (Debian 4.9.2-10+deb8u1) 4.9.2, 64-bit")
    'PostgreSQL 9.3.24'
    >>> pg_get_short_version("PostgreSQL 9.1.4 on x86_64-unknown-linux-gnu, compiled by gcc (GCC) 4.8.5 20150623 (Red Hat 4.8.5-39), 64-bit")
    'PostgreSQL 9.1.4'
    >>> pg_get_short_version("PostgreSQL 14devel on x86_64-pc-linux-gnu, compiled by gcc (GCC) 9.3.1 20200408 (Red Hat 9.3.1-2), 64-bit")
    'PostgreSQL 14devel'
    >>> pg_get_short_version("PostgreSQL 13beta1 on x86_64-pc-linux-gnu, compiled by gcc (GCC) 9.3.1 20200408 (Red Hat 9.3.1-2), 64-bit")
    'PostgreSQL 13beta1'
    >>> pg_get_short_version("PostgreSQL 13rc1 on x86_64-pc-linux-gnu, compiled by gcc (GCC) 9.3.1 20200408 (Red Hat 9.3.1-2), 64-bit")
    'PostgreSQL 13rc1'
    >>> pg_get_short_version("PostgreSQL 9.6rc1 on x86_64-pc-linux-gnu, compiled by gcc (GCC) 9.3.1 20200408 (Red Hat 9.3.1-2), 64-bit")
    'PostgreSQL 9.6rc1'
    """

    res = re.match(
        r"^\w+ [\d\.]+(devel|beta[0-9]+|rc[0-9]+)?",
        text_version,
    )
    if not res:
        raise Exception(f"Undefined PostgreSQL version: {text_version}")

    return res.group(0)


@attr.s(auto_attribs=True, frozen=True, slots=True)
class Data:
    pg_conn: Connection
    pg_version: str
    pg_num_version: int
    server_encoding: bytes
    min_duration: float
    filters: Filters
    dsn_parameters: dict[str, str]
    failed_queries: FailedQueriesInfo

    @classmethod
    def pg_connect(
        cls,
        min_duration: float = 0.0,
        *,
        host: str | None = None,
        port: int = 5432,
        user: str = "postgres",
        password: str | None = None,
        database: str = "postgres",
        rds_mode: bool = False,
        dsn: str = "",
        hide_queries_in_logs: bool = False,
        filters: Filters = NO_FILTER,
    ) -> Data:
        """Create an instance by connecting to a PostgreSQL server."""
        pg_conn = pg.connect(
            dsn,
            host=host,
            port=port,
            user=user,
            dbname=database,
            password=password,
            application_name="pg_activity",
        )
        if hide_queries_in_logs:
            pg.execute(pg_conn, queries.get("disable_log_min_duration_statement"))
            if pg.server_version(pg_conn) >= 130000:
                pg.execute(pg_conn, queries.get("disable_log_min_duration_sample"))
        pg_version = pg_get_short_version(pg_get_version(pg_conn))
        server_encoding = pg_conn.info.parameter_status("server_encoding")
        assert server_encoding is not None
        return cls(
            pg_conn,
            pg_version,
            pg.server_version(pg_conn),
            server_encoding.encode(),
            min_duration=min_duration,
            failed_queries=FailedQueriesInfo(),
            filters=filters,
            dsn_parameters=pg.connection_parameters(pg_conn),
        )

    def try_reconnect(self) -> Data | None:
        try:
            pg_conn = pg.connect(**self.dsn_parameters)
        except (pg.InterfaceError, pg.OperationalError):
            return None
        else:
            return attr.evolve(
                self, pg_conn=pg_conn, dsn_parameters=pg.connection_parameters(pg_conn)
            )

    def pg_is_local_access(self) -> bool:
        """
        Verify if the user running pg_activity can access
        system information for the postmaster process.
        """
        query = queries.get("get_data_directory")
        try:
            ret = pg.fetchone(self.pg_conn, query)
        except pg.InsufficientPrivilege:
            logger.info(
                "Insufficient privilege to show data_directory. System counters are disabled."
            )
            return False

        pid_file = f"{ret['data_directory']}/postmaster.pid"
        try:
            with open(pid_file) as fd:
                pid = fd.readline().strip()
        except OSError as e:
            logger.info(
                "pidfile %s could not be read: %s. System counters are disabled.",
                pid_file,
                e,
            )
            return False

        try:
            proc = psutil.Process(int(pid))
            proc.io_counters()
            proc.cpu_times()
            return True
        except psutil.AccessDenied:
            logger.info(
                "Access denied to the psutil data. System counters are disabled."
            )
            return False
        except AttributeError:
            # See issue #300
            logger.info(
                "Your platform doesn't support some of psutil features required to get system counters. "
                "System counters are disabled."
            )
            return False

    def pg_cancel_backend(self, pid: int) -> bool:
        """
        Cancel a backend
        """
        query = queries.get("do_pg_cancel_backend")
        ret = pg.fetchone(self.pg_conn, query, {"pid": pid})
        return ret["is_stopped"]  # type: ignore[no-any-return]

    def pg_terminate_backend(self, pid: int) -> bool:
        """
        Terminate a backend
        """
        if self.pg_num_version >= 80400:
            query = queries.get("do_pg_terminate_backend")
        else:
            query = queries.get("do_pg_cancel_backend")
        ret = pg.fetchone(self.pg_conn, query, {"pid": pid})
        return ret["is_stopped"]  # type: ignore[no-any-return]

    def pg_get_temporary_file(self) -> TempFileInfo | None:
        """
        Count the number of temporary files and get their total size
        """
        if self.failed_queries.temp_file_query_failed:
            # prevent a spam of errors in PostgreSQL logs if we already failed once
            # for lack of privilege or timeout
            return None

        if self.pg_num_version >= 120000:
            query = queries.get("get_temporary_files_post_120000")
        elif self.pg_num_version >= 90100:
            query = queries.get("get_temporary_files_post_090100")
        else:
            query = queries.get("get_temporary_files_oldest")

        try:
            pg.execute(
                self.pg_conn,
                sql.SQL("SET statement_timeout TO {}").format(sql.Literal("400ms")),
            )
            return pg.fetchone(self.pg_conn, query, mkrow=TempFileInfo)
        except pg.InsufficientPrivilege:
            # superuser or pg_read_server_files are required (Issue #278)
            self.failed_queries.temp_file_query_failed = True
            logger.info(
                "Insufficient privilege to fetch the tempfile data. "
                "The feature was disabled. Please use --no-tempfiles or a platform specific setting (eg. --rds)."
            )

            return None
        except pg.QueryCanceled:
            # if an excessive amount of tempfile exists, the query could be very long
            # to avoid such a case we set a statement_timeout shorter than the lowest
            # refresh rate. This could end up spamming the PostgreSQL logs.
            self.failed_queries.temp_file_query_failed = True
            logger.info(
                "The tempfile query ended in a timeout. "
                "The feature was disabled. Check the temporary files on the server."
            )
            return None
        finally:
            pg.execute(self.pg_conn, queries.get("reset_statement_timeout"))

    def pg_get_wal_senders(self) -> int | None:
        """
        Count the number of wal senders
        """
        if self.pg_num_version >= 90100:
            query = queries.get("get_wal_senders_post_090100")
        else:
            return None
        ret = pg.fetchone(self.pg_conn, query)
        return int(ret["wal_senders"])

    def pg_get_wal_receivers(self) -> int | None:
        """
        Count the number of wal receivers
        """
        if self.failed_queries.wal_receivers_query_failed:
            # prevent a spam of errors un PostgreSQL logs if we already failed once
            # because the feature is not implemented
            return None

        if self.pg_num_version >= 90600:
            query = queries.get("get_wal_receivers_post_090600")
        else:
            return None

        try:
            ret = pg.fetchone(self.pg_conn, query)
        except pg.FeatureNotSupported:
            # Not implemented on Aurora (Issue #301)
            self.failed_queries.wal_receivers_query_failed = True
            logger.info(
                "The receiver information is not available on your platform. "
                "The feature is disabled. Please use the --no-walreceiver option."
            )
            return None

        return int(ret["wal_receivers"])

    def pg_get_replication_slots(self) -> int | None:
        """
        Count the number of replication slots
        """
        if self.pg_num_version >= 140000:
            query = queries.get("get_replication_slots_post_140000")
        else:
            return None
        ret = pg.fetchone(self.pg_conn, query)
        return int(ret["replication_slots"])

    @property
    def dbname_filter(self) -> sql.Composable:
        if self.filters.dbname:
            return sql.Literal(self.filters.dbname)
        return sql.NULL

    def pg_get_server_information(
        self,
        prev_server_info: ServerInformation | None = None,
        using_rds: bool = False,
        skip_db_size: bool = False,
        skip_tempfile: bool = False,
        skip_walreceiver: bool = False,
    ) -> ServerInformation:
        """
        Get the server information (session, workers, cache hit ratio etc..)
        """

        prev_total_size = 0
        if prev_server_info is not None:
            prev_total_size = prev_server_info.total_size

        if self.pg_num_version >= 110000:
            query = queries.get("get_server_info_post_110000")
        elif self.pg_num_version >= 100000:
            query = queries.get("get_server_info_post_100000")
        elif self.pg_num_version >= 90600:
            query = queries.get("get_server_info_post_090600")
        elif self.pg_num_version >= 90400:
            query = queries.get("get_server_info_post_090400")
        elif self.pg_num_version >= 90200:
            query = queries.get("get_server_info_post_090200")
        elif self.pg_num_version >= 90000:
            query = queries.get("get_server_info_post_090000")
        else:
            query = queries.get("get_server_info_oldest")

        qs = sql.SQL(query).format(dbname_filter=self.dbname_filter)
        try:
            ret = pg.fetchone(
                self.pg_conn,
                qs,
                {
                    "dbname_filter": self.filters.dbname,
                    "skip_db_size": skip_db_size,
                    "prev_total_size": prev_total_size,
                    "using_rds": using_rds,
                },
            )
        except pg.InsufficientPrivilege:
            logger.info(
                "Privileges might be insufficient to connect to a database. "
                "Try to use a --filter, the --no-db-size option or a platform specific setting (eg. --rds)"
            )
            raise

        temporary_file_info: TempFileInfo | None = None
        if not skip_tempfile:
            temporary_file_info = self.pg_get_temporary_file()
        wal_senders = self.pg_get_wal_senders()
        wal_receivers: int | None = None
        if not skip_walreceiver:
            wal_receivers = self.pg_get_wal_receivers()
        replication_slots = self.pg_get_replication_slots()

        hr: Pct | None = None
        rr: Pct | None = None
        tps, ips, ups, dps, rps = 0, 0, 0, 0, 0
        size_ev = 0.0
        if prev_server_info is not None:
            dt = float(ret["epoch"] - prev_server_info.epoch)
            try:
                # Note: All this could be negative just after a stat reset
                tps = int((ret["xact_count"] - prev_server_info.xact_count) / dt)
                size_ev = float(ret["total_size"] - prev_server_info.total_size) / dt
                ips = int((ret["insert"] - prev_server_info.insert) / dt)
                ups = int((ret["update"] - prev_server_info.update) / dt)
                dps = int((ret["delete"] - prev_server_info.delete) / dt)
                rps = int(
                    (ret["tuples_returned"] - prev_server_info.tuples_returned) / dt
                )
                deltaread = ret["blks_read"] - prev_server_info.blks_read
                deltahit = ret["blks_hit"] - prev_server_info.blks_hit
                if deltaread + deltahit != 0:
                    hr = Pct(100 * deltahit / (deltaread + deltahit))
                deltacommit = ret["xact_commit"] - prev_server_info.xact_commit
                deltarollback = ret["xact_rollback"] - prev_server_info.xact_rollback
                if deltacommit + deltahit > 0:
                    rr = Pct(100 * deltarollback / (deltacommit + deltarollback))
            except ZeroDivisionError:
                pass

        return ServerInformation(
            size_evolution=size_ev,
            tps=tps,
            insert_per_second=ips,
            update_per_second=ups,
            delete_per_second=dps,
            tuples_returned_per_second=rps,
            cache_hit_ratio_last_snap=hr,
            rollback_ratio_last_snap=rr,
            temporary_file=temporary_file_info,
            wal_senders=wal_senders,
            wal_receivers=wal_receivers,
            replication_slots=replication_slots,
            **ret,
        )

    def pg_get_activities(self, duration_mode: int = 1) -> list[RunningProcess]:
        """
        Get activity from pg_stat_activity view.
        """
        if self.pg_num_version >= 130000:
            qs = queries.get("get_pg_activity_post_130000")
        elif self.pg_num_version >= 110000:
            qs = queries.get("get_pg_activity_post_110000")
        elif self.pg_num_version >= 100000:
            qs = queries.get("get_pg_activity_post_100000")
        elif self.pg_num_version >= 90600:
            qs = queries.get("get_pg_activity_post_090600")
        elif self.pg_num_version >= 90200:
            qs = queries.get("get_pg_activity_post_090200")
        else:
            qs = queries.get("get_pg_activity_oldest")

        duration_column = self.get_duration_column(duration_mode)
        query = sql.SQL(qs).format(
            dbname_filter=self.dbname_filter,
            duration_column=sql.Identifier(duration_column),
            min_duration=sql.Literal(self.min_duration),
        )

        return pg.fetchall(
            self.pg_conn,
            query,
            {
                "min_duration": self.min_duration,
                "dbname_filter": self.filters.dbname,
            },
            mkrow=partial(RunningProcess.from_bytes, self.server_encoding),
            text_as_bytes=True,
        )

    def pg_get_waiting(self, duration_mode: int = 1) -> list[WaitingProcess]:
        """
        Get waiting queries.
        """
        if self.pg_num_version >= 90200:
            qs = queries.get("get_waiting_post_090200")
        else:
            qs = queries.get("get_waiting_oldest")

        duration_column = self.get_duration_column(duration_mode)
        query = sql.SQL(qs).format(
            dbname_filter=self.dbname_filter,
            duration_column=sql.Identifier(duration_column),
            min_duration=sql.Literal(self.min_duration),
        )

        return pg.fetchall(
            self.pg_conn,
            query,
            {
                "min_duration": self.min_duration,
                "dbname_filter": self.filters.dbname,
            },
            mkrow=partial(WaitingProcess.from_bytes, self.server_encoding),
            text_as_bytes=True,
        )

    def pg_get_blocking(self, duration_mode: int = 1) -> list[BlockingProcess]:
        """
        Get blocking queries
        """
        if self.pg_num_version >= 90600:
            qs = queries.get("get_blocking_post_090600")
        elif self.pg_num_version >= 90200:
            qs = queries.get("get_blocking_post_090200")
        else:
            qs = queries.get("get_blocking_oldest")

        duration_column = self.get_duration_column(duration_mode)
        query = sql.SQL(qs).format(
            dbname_filter=self.dbname_filter,
            duration_column=sql.Identifier(duration_column),
            min_duration=sql.Literal(self.min_duration),
        )

        return pg.fetchall(
            self.pg_conn,
            query,
            {
                "min_duration": self.min_duration,
                "dbname_filter": self.filters.dbname,
            },
            mkrow=partial(BlockingProcess.from_bytes, self.server_encoding),
            text_as_bytes=True,
        )

    def pg_is_local(self) -> bool:
        """
        Is pg_activity connected locally?
        """
        query = queries.get("get_pga_inet_addresses")
        ret = pg.fetchone(self.pg_conn, query)
        if ret["inet_server_addr"] == ret["inet_client_addr"]:
            return True
        return False

    @staticmethod
    def get_duration_column(duration_mode: int = 1) -> str:
        """Return the duration column depending on duration_mode.

        >>> Data.get_duration_column(1)
        'query_start'
        >>> Data.get_duration_column(2)
        'xact_start'
        >>> Data.get_duration_column(3)
        'backend_start'
        >>> Data.get_duration_column(9)
        'query_start'
        """
        if duration_mode not in (1, 2, 3):
            duration_mode = 1
        return ["query_start", "xact_start", "backend_start"][duration_mode - 1]


def pg_connect(
    options: Namespace,
    min_duration: float = 0.0,
    filters: Filters = NO_FILTER,
) -> Data:
    """Try to build a Data instance by to connecting to postgres."""
    password = None
    for nb_try in range(2):
        try:
            data = Data.pg_connect(
                dsn=options.connection_string,
                host=options.host,
                port=options.port,
                user=options.username,
                password=password,
                database=options.dbname,
                rds_mode=options.rds,
                min_duration=min_duration,
                filters=filters,
                hide_queries_in_logs=options.hide_queries_in_logs,
            )
        except pg.OperationalError as err:
            if nb_try < 1 and pg.needs_password(err):
                password = getpass.getpass()
            else:
                raise
        except pg.ProgrammingError as err:
            errmsg = str(err).strip()
            if errmsg.startswith("invalid dsn"):
                raise SystemExit(
                    f"ERROR: {errmsg}\n"
                    "Please refer to the 'Connection Control Functions' section of the PostgreSQL documentation"
                )
            raise
        else:
            break
    return data



from __future__ import annotations

import argparse
import logging
import os
import socket
import sys
import time
from io import StringIO
from typing import Any

from blessed import Terminal

from . import __version__, data, types, ui
from .config import Configuration, ConfigurationError
from .pg import OperationalError


def configure_logger(debug_file: str | None = None) -> StringIO:
    logger = logging.getLogger("pgactivity")
    logger.setLevel(logging.DEBUG)

    # The steamhandler is used to print all messages at exit.
    memory_string = StringIO()
    c_handler = logging.StreamHandler(memory_string)
    c_handler.setLevel(logging.INFO)
    c_handler.name = "stream_handler"

    c_format = logging.Formatter("%(levelname)s - %(message)s")
    c_handler.setFormatter(c_format)

    logger.addHandler(c_handler)

    if debug_file is not None:
        f_handler = logging.FileHandler(debug_file)
        f_handler.setLevel(logging.DEBUG)

        f_format = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        f_handler.setFormatter(f_format)

        logger.addHandler(f_handler)

    return memory_string


def flag(p: Any, spec: str, *, dest: str, feature: str) -> None:
    assert not spec.startswith("--no-") and spec.startswith("--"), spec
    p.add_argument(
        spec,
        dest=dest,
        help=f"Enable/disable {feature}.",
        action=argparse.BooleanOptionalAction,
        default=None,
    )


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        usage="%(prog)s [options] [connection string]",
        description=(
            "htop like application for PostgreSQL server activity monitoring."
        ),
        epilog=(
            "The connection string can be in the form of a list of "
            "Key/Value parameters or an URI as described in the PostgreSQL documentation. "
            "The parsing is delegated to the libpq: different versions of the client library "
            "may support different formats or parameters (for example, connection URIs are "
            "only supported from libpq 9.2)."
        ),
        add_help=False,
    )

    group = parser.add_argument_group(
        "Configuration",
    )
    group.add_argument(
        "-P",
        "--profile",
        help=(
            "Configuration profile matching a PROFILE.conf file in "
            "${XDG_CONFIG_HOME:~/.config}/pg_activity/ or /etc/pg_activity/, "
            "or a built-in profile."
        ),
    )

    group = parser.add_argument_group(
        "Options",
    )
    group.add_argument(
        "--blocksize",
        dest="blocksize",
        help="Filesystem blocksize (default: %(default)s).",
        metavar="BLOCKSIZE",
        type=int,
        default=4096,
    )
    group.add_argument(
        "--rds",
        dest="rds",
        action="store_true",
        help="Enable support for AWS RDS (implies --no-tempfiles and filters out the rdsadmin database from space calculation).",
        default=False,
    )
    group.add_argument(
        "--output",
        dest="output",
        help="Store running queries as CSV.",
        metavar="FILEPATH",
        default=None,
    )
    flag(group, "--db-size", dest="dbsize", feature="total size of DB")
    flag(group, "--tempfiles", dest="tempfiles", feature="tempfile count and size")
    flag(group, "--walreceiver", dest="walreceiver", feature="walreceiver checks")
    group.add_argument(
        "-w",
        "--wrap-query",
        dest="wrap_query",
        action="store_true",
        help="Wrap query column instead of truncating.",
        default=False,
    )
    group.add_argument(
        "--duration-mode",
        dest="durationmode",
        help="Duration mode. Values: 1-QUERY(default), 2-TRANSACTION, 3-BACKEND.",
        metavar="DURATION_MODE",
        choices=["1", "2", "3"],
        default="1",
    )
    group.add_argument(
        "--min-duration",
        dest="minduration",
        help="Don't display queries with smaller than specified duration (in seconds).",
        metavar="SECONDS",
        type=float,
        default=0,
    )
    group.add_argument(
        "--filter",
        dest="filters",
        help=(
            "Filter activities with a (case insensitive) regular expression applied on selected fields. "
            "Known fields are: dbname."
        ),
        action="append",
        metavar="FIELD:REGEX",
        default=[],
    )
    group.add_argument(
        "--debug-file",
        dest="debug_file",
        metavar="DEBUG_FILE",
        help="Enable debug and write it to DEBUG_FILE.",
        default=None,
    )
    group.add_argument(
        "--version",
        help="show program's version number and exit.",
        action="version",
        version=f"%(prog)s {__version__}",
    )
    group.add_argument(
        "--help",
        dest="help",
        action="store_true",
        help="Show this help message and exit.",
        default=False,
    )

    group = parser.add_argument_group(
        "Connection Options",
    )
    # Connection string
    group.add_argument(
        "connection_string",
        help=(
            "A valid connection string to the database, e.g.: "
            "'host=HOSTNAME port=PORT user=USER dbname=DBNAME'."
        ),
        nargs="?",
        default="",
        metavar="connection string",
    )
    # -h / --host
    group.add_argument(
        "-h",
        "--host",
        dest="host",
        help="Database server host or socket directory.",
        metavar="HOSTNAME",
    )
    # -p / --port
    group.add_argument(
        "-p",
        "--port",
        dest="port",
        help="Database server port.",
        metavar="PORT",
    )
    # -U / --username
    group.add_argument(
        "-U",
        "--username",
        dest="username",
        help="Database user name.",
        metavar="USERNAME",
    )
    # -d / --dbname
    group.add_argument(
        "-d",
        "--dbname",
        dest="dbname",
        help="Database name to connect to.",
        metavar="DBNAME",
    )

    group = parser.add_argument_group(
        "Process table display options",
        "These options may be used hide some columns from the processes table.",
    )
    flag(group, "--pid", dest="pid", feature="PID")
    flag(group, "--xmin", dest="xmin", feature="XMIN")
    flag(group, "--database", dest="database", feature="DATABASE")
    flag(group, "--user", dest="user", feature="USER")
    flag(group, "--client", dest="client", feature="CLIENT")
    flag(group, "--cpu", dest="cpu", feature="CPU%%")
    flag(group, "--mem", dest="mem", feature="MEM%%")
    flag(group, "--read", dest="read", feature="READ/s")
    flag(group, "--write", dest="write", feature="WRITE/s")
    flag(group, "--time", dest="time", feature="TIME+")
    flag(group, "--wait", dest="wait", feature="W")
    flag(group, "--app-name", dest="appname", feature="APP")

    group = parser.add_argument_group("Header display options")
    group.add_argument(
        "--no-inst-info",
        dest="header_show_instance",
        action="store_false",
        help="Hide instance information.",
        default=None,
    )
    group.add_argument(
        "--no-sys-info",
        dest="header_show_system",
        action="store_false",
        help="Hide system information.",
        default=None,
    )
    group.add_argument(
        "--no-proc-info",
        dest="header_show_workers",
        action="store_false",
        help="Hide workers process information.",
        default=None,
    )

    group = parser.add_argument_group("Other display options")
    group.add_argument(
        "--hide-queries-in-logs",
        dest="hide_queries_in_logs",
        action="store_true",
        help="Disable log_min_duration_statements and log_min_duration_sample for pg_activity.",
        default=False,
    )
    group.add_argument(
        "--refresh",
        dest="refresh",
        help="Refresh rate. Values: %(choices)s (default: %(default)d).",
        metavar="REFRESH",
        choices=[0.5, 1, 2, 3, 4, 5],
        type=float,
        default=2,
    )

    return parser


def main() -> None:
    if os.name != "posix":
        sys.exit("FATAL: Platform not supported.")

    parser = get_parser()
    args = parser.parse_args()
    memory_stream = configure_logger(args.debug_file)

    if args.help:
        parser.print_help()
        sys.exit(1)

    try:
        filters = types.Filters.from_options(args.filters)
    except ValueError as e:
        parser.error(str(e))
    if args.rds:
        args.notempfile = True

    try:
        cfg = Configuration.lookup(args.profile)
    except (ConfigurationError, FileNotFoundError) as e:
        parser.error(str(e))

    try:
        dataobj = data.pg_connect(args, min_duration=args.minduration, filters=filters)
    except OperationalError as e:
        parser.exit(status=1, message=f"could not connect to PostgreSQL: {e}")
    hostname = socket.gethostname()
    conninfo = dataobj.pg_conn.info
    host = types.Host(
        hostname,
        conninfo.user,
        conninfo.host,
        int(conninfo.port),
        conninfo.dbname,
    )

    term = Terminal()
    while True:
        try:
            ui.main(term, cfg, dataobj, host, args)
        except ConfigurationError as e:
            parser.exit(1, f"error: {e}")
        except OperationalError:
            while True:
                print(term.clear + term.home, end="")
                print("Connection to PostgreSQL lost, trying to reconnect...")
                try:
                    time.sleep(5)
                except KeyboardInterrupt:
                    sys.exit(1)
                newdataobj = dataobj.try_reconnect()
                if newdataobj is not None:
                    dataobj = newdataobj
                    break
        except KeyboardInterrupt:
            sys.exit(0)
        else:
            break
        finally:
            print(memory_stream.getvalue(), file=sys.stderr)
